#!/usr/bin/env python

import logging
l = logging.getLogger("claripy.frontends.light_frontend")

from .constrained_frontend import ConstrainedFrontend

class LightFrontend(ConstrainedFrontend):
    def __init__(self, solver_backend, cache=None, **kwargs):
        # since the LightFrontend mostly cannot handle extra_constriants, it horribly misuses the cache.
        # Because of this, we have to disable the caching here.
        ConstrainedFrontend.__init__(self, cache=False if cache is None else cache, **kwargs)
        self._solver_backend = solver_backend

    #
    # Storable support
    #

    def _ana_getstate(self):
        return self._solver_backend.__class__.__name__, ConstrainedFrontend._ana_getstate(self)

    def _ana_setstate(self, s):
        solver_backend_name, base_state = s
        ConstrainedFrontend._ana_setstate(self, base_state)
        self._solver_backend = backends._backends_by_type[solver_backend_name]

    #
    # Light functionality
    #

    def _eval(self, e, n, extra_constraints=(), exact=None, cache=None):
        try:
            return self._solver_backend.eval(e, n, result=self.result)
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this eval().")

    def _batch_eval(self, exprs, n, extra_constraints=(), exact=None, cache=None):
        try:
            return self._solver_backend._batch_eval(exprs, n, result=self.result)
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this batch_eval().")

    def _max(self, e, extra_constraints=(), exact=None, cache=None):
        try:
            return self._solver_backend.max(e, result=self.result)
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this max().")

    def _min(self, e, extra_constraints=(), exact=None, cache=None):
        try:
            return self._solver_backend.min(e, result=self.result)
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this min().")

    def _solution(self, e, v, extra_constraints=(), exact=None, cache=None):
        try:
            return self._solver_backend.solution(e, v, result=self.result)
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this solution().")

    def _is_true(self, e, extra_constraints=(), exact=None, cache=None):
        try:
            return self._solver_backend.is_true(e, result=self.result)
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this is_false().")

    def _is_false(self, e, extra_constraints=(), exact=None, cache=None):
        try:
            return self._solver_backend.is_false(e, result=self.result)
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this is_false().")

    def _satisfiable(self, extra_constraints=(), exact=None, cache=None):
        try:
            return any(self._solver_backend.is_false(c) for c in reversed(self.constraints + list(extra_constraints)))
        except BackendError:
            raise ClaripyFrontendError("Light solver can't handle this is_false().")

    #
    # Merging and splitting
    #

    def _blank_copy(self):
        return LightFrontend(self._solver_backend, cache=self._cache)

    def merge(self, others, merge_flag, merge_values):
        return self._solver_backend.__class__.__name__ == 'BackendZ3', ConstrainedFrontend.merge(self, others, merge_flag, merge_values)[1]

from ..errors import BackendError, ClaripyFrontendError
from ..backend_manager import backends



#!/usr/bin/env python

import weakref
import logging
l = logging.getLogger("claripy.frontends.full_frontend")

from .constrained_frontend import ConstrainedFrontend

class ReplacementFrontend(ConstrainedFrontend):
    def __init__(self, actual_frontend, allow_symbolic=None, replacements=None, replacement_cache=None, unsafe_replacement=None, auto_replace=None, replace_constraints=None, **kwargs):
        kwargs['cache'] = kwargs.get('cache', False)
        ConstrainedFrontend.__init__(self, **kwargs)
        self._actual_frontend = actual_frontend
        self._allow_symbolic = True if allow_symbolic is None else allow_symbolic
        self._auto_replace = True if auto_replace is None else auto_replace
        self._replace_constraints = False if replace_constraints is None else replace_constraints
        self._unsafe_replacement = False if unsafe_replacement is None else unsafe_replacement
        self._replacements = { } if replacements is None else replacements
        self._replacement_cache = weakref.WeakKeyDictionary() if replacement_cache is None else replacement_cache

    def add_replacement(self, old, new, invalidate_cache=True, replace=True):
        if not isinstance(old, Base):
            return

        if not replace and old in self._replacements:
            return

        if not isinstance(new, Base):
            if isinstance(new, bool):
                new = BoolV(new)
            elif isinstance(new, (int, long)):
                new = BVV(new, old.length)
            else:
                return

        if invalidate_cache:
            self._replacements = dict(self._replacements)
            self._replacement_cache = weakref.WeakKeyDictionary(self._replacement_cache)

        self._replacements[old.cache_key] = new
        self._replacement_cache[old.cache_key] = new

    def _replacement(self, old):
        if not isinstance(old, Base):
            return old

        if old.cache_key in self._replacement_cache:
            return self._replacement_cache[old.cache_key]
        else:
            new = old.replace_dict(self._replacement_cache)
            if new is not old:
                self._replacement_cache[old.cache_key] = new
            return new

    def _add_solve_result(self, e, er, r):
        if not self._auto_replace:
            return
        if not isinstance(e, Base) or not e.symbolic:
            return
        if er.symbolic:
            return
        self.add_replacement(e, r, invalidate_cache=False)


    #
    # Storable support
    #

    def _blank_copy(self):
        s = ReplacementFrontend(self._actual_frontend._blank_copy())
        s._auto_replace = self._auto_replace
        s._replace_constraints = self._replace_constraints
        s._unsafe_replacement = self._unsafe_replacement
        s._allow_symbolic = self._allow_symbolic
        return s

    def branch(self):
        s = ConstrainedFrontend.branch(self)
        s._action_frontend = self._actual_frontend.branch()
        s._replacements = self._replacements
        s._replacement_cache = self._replacement_cache
        return s

    def downsize(self):
        self._actual_frontend.downsize()
        self._replacement_cache.clear()

    def _ana_getstate(self):
        return self._replacements, self._actual_frontend, ConstrainedFrontend._ana_getstate(self)

    def _ana_setstate(self, s):
        self._replacements, self._actual_frontend, base_state = s
        ConstrainedFrontend._ana_setstate(self, base_state)
        self._replacement_cache = weakref.WeakKeyDictionary()

    #
    # Replacement solving
    #

    def _replace_list(self, lst):
        return tuple(self._replacement(c) for c in lst)

    def solve(self, extra_constraints=(), exact=None, cache=None):
        ecr = self._replace_list(extra_constraints)
        return self._actual_frontend.solve(extra_constraints=ecr, exact=exact, cache=cache)

    def eval(self, e, n, extra_constraints=(), exact=None, cache=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.eval(er, n, extra_constraints=ecr, exact=exact, cache=cache)
        if self._unsafe_replacement: self._add_solve_result(e, er, r[0])
        return r

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None, cache=None):
        er = self._replace_list(exprs)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.batch_eval(er, n, extra_constraints=ecr, exact=exact, cache=cache)
        if self._unsafe_replacement:
            for i, original in enumerate(exprs):
                self._add_solve_result(original, er[i], r[0][i])
        return r

    def max(self, e, extra_constraints=(), exact=None, cache=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.max(er, extra_constraints=ecr, exact=exact, cache=cache)
        if self._unsafe_replacement: self._add_solve_result(e, er, r)
        return r

    def min(self, e, extra_constraints=(), exact=None, cache=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.min(er, extra_constraints=ecr, exact=exact, cache=cache)
        if self._unsafe_replacement: self._add_solve_result(e, er, r)
        return r

    def solution(self, e, v, extra_constraints=(), exact=None, cache=None):
        er = self._replacement(e)
        vr = self._replacement(v)
        ecr = self._replace_list(extra_constraints)
        r = self._actual_frontend.solution(er, vr, extra_constraints=ecr, exact=exact, cache=cache)
        if self._unsafe_replacement and r and (not isinstance(vr, Base) or not vr.symbolic):
            self._add_solve_result(e, er, vr)
        return r

    def is_true(self, e, extra_constraints=(), exact=None, cache=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        return self._actual_frontend.is_true(er, extra_constraints=ecr, exact=exact, cache=cache)

    def is_false(self, e, extra_constraints=(), exact=None, cache=None):
        er = self._replacement(e)
        ecr = self._replace_list(extra_constraints)
        return self._actual_frontend.is_false(er, extra_constraints=ecr, exact=exact, cache=cache)

    def satisfiable(self, extra_constraints=(), exact=None, cache=None):
        ecr = self._replace_list(extra_constraints)
        return self._actual_frontend.satisfiable(extra_constraints=ecr, exact=exact, cache=cache)

    def _filter_single_constraint(self, e):
        if self._replace_constraints:
            er = self._replacement(e)
            return super(ReplacementFrontend, self)._filter_single_constraint(er)
        else:
            return super(ReplacementFrontend, self)._filter_single_constraint(e)

    def _add_constraints(self, constraints, **kwargs):
        for c in constraints:
            if self._auto_replace and isinstance(c, Base) and c.op == '__eq__' and isinstance(c.args[0], Base) and isinstance(c.args[1], Base):
                #print c; import ipdb; ipdb.set_trace()
                if c.args[0].symbolic ^ c.args[1].symbolic:
                    if c.args[0].cache_key not in self._replacements and c.args[0].cache_key not in self._replacement_cache:
                        self.add_replacement(c.args[0], c.args[1], invalidate_cache=True)
                    elif c.args[1].cache_key not in self._replacements and c.args[1].cache_key not in self._replacement_cache:
                        self.add_replacement(c.args[1], c.args[0], invalidate_cache=True)

        ConstrainedFrontend._add_constraints(self, constraints, **kwargs)

        cr = self._replace_list(constraints)
        if not self._allow_symbolic and any(c.symbolic for c in cr):
            raise ClaripyFrontendError("symbolic constraints made it into ReplacementFrontend with allow_symbolic=False")
        return self._actual_frontend.add(cr, **kwargs)

    #def _add_constraints(self, *args, **kwargs): #pylint:disable=unused-argument
    #   raise Exception("this should not be called")
    def _solve(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")
    def _eval(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")
    def _batch_eval(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")
    def _max(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")
    def _min(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")
    def _solution(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")
    def _is_true(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")
    def _is_false(self, *args, **kwargs): #pylint:disable=unused-argument
        raise Exception("this should not be called")

from ..ast.base import Base
from ..ast.bv import BVV
from ..ast.bool import BoolV
from ..errors import ClaripyFrontendError



import logging

l = logging.getLogger('claripy.balancer')

class Balancer(object):
    '''
    The Balancer is an equation redistributor. The idea is to take an AST and
    rebalance it to, for example, isolate unknown terms on one side of an
    inequality.
    '''

    def __init__(self, backend):
        self._backend = backend

    def constraint_to_si(self, expr):
        """
        We take in constraints, and convert them into constrained strided-intervals.

        For example, expr =
        <A __ne__ (
                    <A Extract (0,
                                0,
                                <A Concat (<A BVV(0x0, 63)>,
                                            <A If (
                                                    <A ULE (<A SI_1208<64>0x1[0x0, 0x100]>, <A BVV(0x27, 64)>)>,
                                                    <A BVV(0x1, 1)>,
                                                    <A BVV(0x0, 1)>
                                            )>
                                )>
                    )>,
                    0
        )>
        The result would be
        [ ( SI_1208<64>0x1[0x0, 0x100], SI_XXXX<64>0x1[0x0, 0x27] ) ]

        As we can only deal with bits, we will convert all integers into BVV during the solving and conversion process.

        :param expr: The constraint
        :return: whether the expr is satisfiable (boolean), and a list of tuples in form of
                (original_si, constrained_si).
        """

        try:
            sat, lst = self._handle(expr.op, expr.args)
            return sat, lst

        except ClaripyBalancerError as ex:
            l.error('VSASimplifiers raised an exception %s. Please report it.', str(ex), exc_info=True)

            # return the dummy result
            return True, [ ]

    def _simplify(self, op, args, expr, condition):
        handler_name = "_simplify_%s" % op
        if not hasattr(self, handler_name):
            l.error('Simplification handler "%s" is not found in balancer. Consider implementing.', handler_name)
            return expr, condition

        new_expr, new_cond = getattr(self, "_simplify_%s" % op)(args, expr, condition)
        return new_expr, new_cond

    def _handle(self, op, args):

        if len(args) == 2:
            lhs, rhs = args

            # Simplify left side
            lhs, new_cond = self._simplify(lhs.op, lhs.args, lhs, (op, rhs))

            # Update args
            op, rhs = new_cond
            args = (lhs, rhs)

            sat, lst = getattr(self, "_handle_%s" % op)(args)

        else:
            sat, lst = getattr(self, "_handle_%s" % op)(args)

        return sat, lst

    #
    # Dealing with constraints
    #

    reversed_operations = { }
    reversed_operations['__ne__'] = '__eq__'
    reversed_operations['__eq__'] = '__ne__'
    reversed_operations['ULT'] = 'UGE'
    reversed_operations['UGT'] = 'ULE'
    reversed_operations['ULE'] = 'UGT'
    reversed_operations['UGE'] = 'ULT'
    reversed_operations['SLT'] = 'SGE'
    reversed_operations['SLE'] = 'SGT'
    reversed_operations['SGT'] = 'SLE'
    reversed_operations['SGE'] = 'SLT'
    reversed_operations['__le__'] = '__gt__'
    reversed_operations['__lt__'] = '__ge__'
    reversed_operations['__ge__'] = '__lt__'
    reversed_operations['__gt__'] = '__le__'

    comparison_info = { }
    # Tuples look like (is_lt, is_eq, is_unsigned)
    comparison_info['SLT'] = (True, False, False)
    comparison_info['SLE'] = (True, True, False)
    comparison_info['SGT'] = (False, False, False)
    comparison_info['SGE'] = (False, True, False)
    comparison_info['ULT'] = (True, False, True)
    comparison_info['ULE'] = (True, True, True)
    comparison_info['UGT'] = (False, False, True)
    comparison_info['UGE'] = (False, True, True)
    comparison_info['__lt__'] = comparison_info['ULT']
    comparison_info['__le__'] = comparison_info['ULE']
    comparison_info['__gt__'] = comparison_info['UGT']
    comparison_info['__ge__'] = comparison_info['UGE']

    def _simplify_ZeroExt(self, args, expr, condition):
        """
        :param args:
        :param expr:
        :return: (new ast, new condition)
        """
        cond_op, cond_arg = condition
        # Normalize cond_arg
        if type(cond_arg) in (int, long): #pylint:disable=unidiomatic-typecheck
            cond_arg = _all_operations.BVV(cond_arg, expr.size())

        extended_bits, arg = args

        if cond_arg.size() <= arg.size() or \
                is_true(cond_arg[ expr.size() - 1 : expr.size() - extended_bits ] == 0):
            # We can safely eliminate this layer of ZeroExt
            if cond_arg.size() < arg.size():
                larger_cond_arg = cond_arg.zero_extend(arg.size() - cond_arg.size())
                if not isinstance(self._backend.convert(larger_cond_arg), Base):
                    return self._simplify(arg.op, arg.args, arg, (cond_op, larger_cond_arg))
            else:
                return self._simplify(arg.op, arg.args, arg, (cond_op, cond_arg[ arg.size() - 1 : 0 ]))

        else:
            # TODO: We may also handle the '__eq__' and '__ne__' case
            # We cannot handle this...
            return expr, condition

    def _simplify_BVS(self, args, expr, condition): #pylint:disable=no-self-use,unused-argument
        return expr, condition

    def _simplify_SignExt(self, args, expr, condition):
        """

        :param args:
        :param expr:
        :param condition:
        :return:
        """
        # TODO: Review the logic of this method
        cond_op, cond_arg = condition
        # Normalize them
        if type(cond_arg) in (int, long): #pylint:disable=unidiomatic-typecheck
            cond_arg = _all_operations.BVV(cond_arg, expr.size())

        extended_bits, arg = args

        if cond_arg.size() <= arg.size() or \
                is_true(cond_arg[expr.size() - 1: expr.size() - extended_bits] == 0):
            # We can safely eliminate this layer of SignExt
            if cond_arg.size() < arg.size():
                larger_cond_arg = cond_arg.zero_extend(arg.size() - cond_arg.size()).resolved()
                if not isinstance(larger_cond_arg, Base):
                    return self._simplify(arg.op, arg.args, arg, (cond_op, larger_cond_arg))
            else:
                return self._simplify(arg.op, arg.args, arg, (cond_op, cond_arg[arg.size() - 1: 0]))

        else:
            # TODO: We may also handle the '__eq__' and '__ne__' case
            # We cannot handle this...
            return expr, condition

    def _simplify_Extract(self, args, expr, condition):
        '''
        Convert Extract(a, b, If(...)) to If(..., Extract(a, b, ...), Extract(a, b, ...))

        :param args:
        :return:
        '''

        high, low, to_extract = args
        cond_operation, cond_operand = condition
        # Make sure the condition operand has the same size as to_extract
        new_condition = cond_operation, _all_operations.ZeroExt((high - low + 1), cond_operand)
        ast, cond = self._simplify(to_extract.op, to_extract.args, to_extract, new_condition)

        # Create the new ifproxy
        if ast is None:
            # We cannot handle it
            return None, condition

        elif ast.op == 'If':
            new_ifproxy = _all_operations.If(
                                ast.args[0],
                                _all_operations.Extract(high, low, ast.args[1]),
                                _all_operations.Extract(high, low, ast.args[2])
                            )

        else:
            cond_op, cond_arg = cond
            if type(self._backend.convert(cond_arg)) in (int, long): #pylint:disable=unidiomatic-typecheck
                cond_arg = _all_operations.BVV(cond_arg, ast.size())
            elif type(self._backend.convert(cond_arg)) in (vsa.StridedInterval, vsa.DiscreteStridedIntervalSet, bv.BVV): #pylint:disable=unidiomatic-typecheck
                if ast.size() > cond_arg.size():
                    # Make sure two operands have the same size
                    cond_arg = _all_operations.ZeroExt(ast.size() - cond_arg.size(), cond_arg)

            if cond_arg.size() - 1 < high + 1 or \
                    is_true(cond_arg[cond_arg.size() - 1 : high + 1] == 0):
                # The upper part doesn't matter
                # We can handle it
                return self._simplify(ast.op, ast.args, ast, (cond_op, cond_arg))
            else:
                # We cannot further simplify it
                return expr, condition

        return new_ifproxy, condition

    def _simplify_Concat(self, args, expr, condition):
        '''
        Convert Concat(a, If(...)) to If(..., Concat(a, ...), Concat(a, ...))

        :param args:
        :return:
        '''

        new_args = [ self._simplify(ex.op, ex.args, ex, condition) for ex in args ]

        ifproxy_conds = set([ a.args[0] for a, new_cond in new_args if a.op == 'If' ])

        if len(ifproxy_conds) == 0:
            # Let's check if we can remove this layer of Concat
            cond = condition[1]
            if len(args) == 2:
                if cond.size() - 1 >= cond.size() - args[0].size():
                    if is_true(args[0] == cond[ cond.size() - 1 : cond.size() - args[0].size() ]):
                        # Yes! We can remove it!
                        # TODO: This is hackish...
                        new_cond = (condition[0], cond[ cond.size() - args[0].size() - 1 : 0])
                        return self._simplify(args[1].op, args[1].args, args[1], new_cond)

                else:
                    # args[0].size() == 0? It must be a bug.
                    raise ClaripyBackendVSAError(
                        'args[0].size() == %d (args[0] is \'%s\'). Please report this bug.' % (args[0].size, str(args[0])))

            # Cannot simplify it anymore
            return expr, condition

        elif len(ifproxy_conds) > 1:
            # We have more than one condition. Cannot handle it for now!
            return None, condition

        else:
            concat_trueexpr = [ ]
            concat_falseexpr = [ ]

            all_new_conds = set([ new_cond for a, new_cond in new_args ])

            if len(all_new_conds) > 1:
                # New conditions are not consistent. Can't handle it.
                return expr, condition

            for a, new_cond in new_args:
                if a.op == "If":
                    concat_trueexpr.append(a.args[1])
                    concat_falseexpr.append(a.args[2])
                else:
                    concat_trueexpr.append(a)
                    concat_falseexpr.append(a)

            new_ifproxy = _all_operations.If(
                                list(ifproxy_conds)[0],
                                _all_operations.Concat(*concat_trueexpr),
                                _all_operations.Concat(*concat_falseexpr)
                            )

            return new_ifproxy, condition

    def _simplify_I(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use
        return expr, condition

    def _simplify_If(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use
        return expr, condition

    def _simplify_Reverse(self, args, expr, condition): #pylint:disable=unused-argument
        # TODO: How should we deal with Reverse in a smart way?

        arg = args[0]

        return self._simplify(arg.op, arg.args, arg, condition)

    def _simplify_widen(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def _simplify_intersection(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def _simplify___or__(self, args, expr, condition):
        claripy = expr._claripy
        argl, argr = args
        if argl is argr or claripy.is_true(argl == argr):
            return self._simplify(argl.op, argl.args, argl, condition)
        elif claripy.is_true(argl == 0):
            return self._simplify(argr.op, argr.args, argr, condition)
        elif claripy.is_true(argr == 0):
            return self._simplify(argl.op, argl.args, argl, condition)
        else:
            return expr, condition

    def _simplify___and__(self, args, expr, condition):

        argl, argr = args
        if argl is argr:
            # Operands are the same one!
            # We can safely remove this layer of __and__
            return self._simplify(argl.op, argl.args, argl, condition)

        elif argl.structurally_match(argr):
            # Operands are the same
            # Safely remove the __and__ operation
            return self._simplify(argl.op, argl.args, argl, condition)

        else:
            # We cannot handle it
            return expr, condition

    def _simplify___xor__(self, args, expr, condition):
        argl, argr = args

        if is_true(argl == 0):
            # :-)
            return self._simplify(argr.op, argr.args, argr, condition)
        elif is_true(argr == 0):
            # :-)
            return self._simplify(argl.op, argl.args, argl, condition)
        else:
            # :-(
            return expr, condition

    def _simplify___add__(self, args, expr, condition):

        argl, argr = args
        if is_true(argr == 0):
            # This layer of __add__ can be removed
            return self._simplify(argl.op, argl.args, argl, condition)
        elif is_true(argl == 0):
            # This layer of __add__ can be removed
            return self._simplify(argr.op, argr.args, argr, condition)
        else:

            if isinstance(self._backend.convert(argl), bv.BVV):
                new_cond = (condition[0], condition[1] - argl)
                return self._simplify(argr.op, argr.args, argr, new_cond)

            elif isinstance(self._backend.convert(argr), bv.BVV):
                new_cond = (condition[0], condition[1] - argr)
                return self._simplify(argl.op, argl.args, argl, new_cond)

            else:
                return expr, condition

    def _simplify___radd__(self, args, expr, condition):
        return self._simplify___add__((args[1], args[0]), expr, condition)

    def _simplify___sub__(self, args, expr, condition):
        """

        :param args:
        :param expr:
        :param condition:
        :return:
        """

        argl, argr = args
        if is_true(argr == 0):
            return self._simplify(argl.op, argl.args, argl, condition)
        elif is_true(argl == 0):
            return self._simplify(argr.op, argr.args, argr, condition)
        else:
            return expr, condition

    def _simplify___rsub__(self, args, expr, condition):
        return self._simplify___sub__((args[1], args[0]), expr, condition)

    def _simplify___rshift__(self, args, expr, condition):

        arg, offset = args
        if is_true(offset == 0):
            return self._simplify(arg.op, arg.args, arg, condition)
        else:
            return expr, condition

    def _simplify___lshift__(self, args, expr, condition):

        arg, offset = args
        if is_true(offset == 0):
            return self._simplify(arg.op, arg.args, arg, condition)
        else:
            return expr, condition

    def _simplify___invert__(self, args, expr, condition):

        arg = args[0]
        if arg.op == 'If':
            new_arg = _all_operations.If(args[0], args[1].__invert__(), args[2].__invert__())

            return self._simplify(new_arg.op, new_arg.args, expr, condition)

        else:
            return expr, condition

    def _simplify_union(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def _simplify___mod__(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def _simplify___div__(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def _simplify___eq__(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        l.error('_simplify___eq__() should not exist. This is just a workaround for VSA. Fish will fix the issue later.')

        return expr, condition

    def _handle_comparison(self, args, comp=None):
        """
        Handles all comparisons.
        :param args:
        :param comp:
        :return:
        """

        if comp in self.comparison_info:
            is_lt, is_equal, is_unsigned = self.comparison_info[comp]
        else:
            raise ClaripyBalancerError('Support for comparison %s is not implemented. Please report it.' % comp)

        lhs, rhs = args

        if not isinstance(lhs, Base):
            raise ClaripyBalancerError('Left-hand-side expression is not an AST object.')

        lhs_bo = self._backend.convert(lhs)
        rhs_bo = self._backend.convert(rhs)

        # Maybe the target variable is the rhs
        if lhs.cardinality == 1 or lhs_bo.is_empty:
            new_op = self.reversed_operations[comp]
            new_lhs, new_rhs = rhs, lhs
            return self._handle(new_op, (new_lhs, new_rhs))

        if lhs.op == 'If':
            condition, trueexpr, falseexpr = lhs.args
            trueexpr = self._backend.convert(trueexpr)
            falseexpr = self._backend.convert(falseexpr)

            if is_unsigned:
                if is_lt:
                    if is_equal:
                        take_true = is_true(trueexpr.ULE(rhs_bo))
                        take_false = is_true(falseexpr.ULE(rhs_bo))
                    else:
                        take_true = is_true(falseexpr.ULT(rhs_bo))
                        take_false = is_true(trueexpr.ULT(rhs_bo))
                else:
                    if is_equal:
                        take_true = is_true(trueexpr.UGE(rhs_bo))
                        take_false = is_true(falseexpr.UGE(rhs_bo))
                    else:
                        take_true = is_true(trueexpr.UGT(rhs_bo))
                        take_false = is_true(falseexpr.UGT(rhs_bo))
            else:
                if is_lt:
                    if is_equal:
                        take_true = is_true(trueexpr <= rhs_bo)
                        take_false = is_true(falseexpr <= rhs_bo)
                    else:
                        take_true = is_true(trueexpr < rhs_bo)
                        take_false = is_true(falseexpr < rhs_bo)
                else:
                    if is_equal:
                        take_true = is_true(trueexpr >= rhs_bo)
                        take_false = is_true(falseexpr >= rhs_bo)
                    else:
                        take_true = is_true(trueexpr > rhs_bo)
                        take_false = is_true(falseexpr > rhs_bo)

            if take_true and take_false:
                # It's always satisfiable
                return True, [ ]
            elif take_true:
                return self._handle(condition.op, condition.args)
            elif take_false:
                rev_op = self.reversed_operations[condition.op]
                return self._handle(rev_op, condition.args)
            else:
                # Not satisfiable
                return False, [ ]

        elif isinstance(rhs_bo, vsa.StridedInterval) and isinstance(lhs_bo, vsa.StridedInterval):
            if isinstance(lhs_bo, Base):
                # It cannot be computed by our backend...
                # We just give up for now
                return True, [ ]

            stride = lhs_bo.stride

            if is_lt:
                # < or <=
                if is_unsigned: lb = 0
                else: lb = vsa.StridedInterval.min_int(rhs.length)

                ub = rhs_bo.upper_bound
                if not is_equal:
                    # <
                    ub = ub - 1

            else:
                # > or >=
                if is_unsigned: ub = vsa.StridedInterval.max_int(rhs.length)
                else: ub = vsa.StridedInterval.max_int(rhs.length - 1)

                lb = rhs_bo.lower_bound
                if not is_equal:
                    # >
                    lb = lb + 1

            if stride == 0 and lb != ub:
                # Make sure the final vsa.StridedInterval is always meaningful. See issue #55.
                stride = 1

            si_replacement = _all_operations.SI(bits=rhs.length, stride=stride, lower_bound=lb, upper_bound=ub)
            return True, [(lhs, si_replacement)]

        else:
            return True, [ ]

    def _handle___lt__(self, args): return self._handle_comparison(args, comp='__lt__')
    def _handle___le__(self, args): return self._handle_comparison(args, comp='__le__')
    def _handle___gt__(self, args): return self._handle_comparison(args, comp='__gt__')
    def _handle___ge__(self, args): return self._handle_comparison(args, comp='__ge__')
    def _handle_ULT(self, args): return self._handle_comparison(args, comp='ULT')
    def _handle_ULE(self, args): return self._handle_comparison(args, comp='ULE')
    def _handle_UGT(self, args): return self._handle_comparison(args, comp='UGT')
    def _handle_UGE(self, args): return self._handle_comparison(args, comp='UGE')
    def _handle_SLT(self, args): return self._handle_comparison(args, comp='SLT')
    def _handle_SLE(self, args): return self._handle_comparison(args, comp='SLE')
    def _handle_SGT(self, args): return self._handle_comparison(args, comp='SGT')
    def _handle_SGE(self, args): return self._handle_comparison(args, comp='SGE')

    def _handle_I(self, args): #pylint:disable=no-self-use
        a = args[0]

        if a in (False, 0):
            return False, [ ]
        elif isinstance(a, bv.BVV) and a.value == 0:
            return False, [ ]

        return True, [ ]

    def _handle_Not(self, args):
        """
        The argument should be False

        :param args:
        :return:
        """

        a = args[0]
        expr_op = a.op
        expr_args = a.args

        # Reverse the op
        expr_op = self.reversed_operations[expr_op]

        return self._handle(expr_op, expr_args)

    def _handle_And(self, args):
        """

        :param args:
        :return:
        """

        sat = True
        lst = [ ]

        # Both sides must be true
        for arg in args:
            sat_, lst_ = self._handle(arg.op, arg.args)

            sat &= sat_
            lst.extend(lst_)

        if not sat:
            lst = [ ]

        return sat, lst

    def _handle_Or(self, args):

        if len(args) == 1:
            return self._handle(args[0].op, args[0].args)

        else:
            if len(args) > 0:
                args = [ self._handle(a.op, a.args) for a in args ]
                if any([not is_false(a) for a in args]):
                    return True, [ ]

                else:
                    return False, [ ]
            return True, [ ]

    def _handle___ne__(self, args):
        return self._handle_eq_ne(args, False)

    def _handle___eq__(self, args):
        return self._handle_eq_ne(args, True)

    def _handle_eq_ne(self, args, is_eq):
        """

        :param args:
        :return: True or False, and a list of (original_si, constrained_si) tuples
        """

        lhs, rhs = args

        if not isinstance(lhs, Base):
            raise ClaripyBalancerError('Left-hand-side expression is not an AST object.')

        size = lhs.size()

        if type(rhs) in (int, long): #pylint:disable=unidiomatic-typecheck
            # Convert it into a BVV
            rhs = _all_operations.BVV(rhs, size)

        if not isinstance(rhs, Base):
            raise ClaripyBalancerError('Right-hand-side expression cannot be converted to an AST object.')

        # TODO: Make sure the rhs doesn't contain any IfProxy

        if lhs.op == 'If':
            condition, trueexpr, falseexpr = lhs.args

            if is_eq:
                # __eq__
                take_true = is_true(rhs == trueexpr)
                take_false = is_true(rhs == falseexpr)
            else:
                # __ne__
                take_true = is_true(rhs == falseexpr)
                take_false = is_true(rhs == trueexpr)

            if take_true and take_false:
                # It's always satisfiable
                return True, [ ]

            elif take_true:
                # We take the true side
                return self._handle(condition.op, condition.args)

            elif take_false:
                # We take the false side

                # Reverse the operation first
                rev_op = self.reversed_operations[condition.op]

                return self._handle(rev_op, condition.args)

            else:
                # Not satisfiable
                return False, [ ]
        elif isinstance(self._backend.convert(lhs), vsa.StridedInterval) or isinstance(self._backend.convert(lhs), bv.BVV):
            if not isinstance(self._backend.convert(lhs), vsa.StridedInterval):
                try: lhs = _all_operations.SI(to_conv=lhs)
                except BackendError: return True, [ ] # We cannot convert it to a vsa.StridedInterval

            try: rhs = self._backend.convert(rhs)
            except BackendError: return True, [ ]

            if is_eq:
                return True, [ (lhs, rhs)]
            else:
                lhs_bo = self._backend.convert(lhs)
                rhs_bo = self._backend.convert(rhs)

                if lhs_bo.upper_bound <= rhs_bo.upper_bound:
                    r = self._backend.CreateStridedInterval(bits=rhs_bo.bits,
                                    stride=lhs_bo.stride,
                                    lower_bound=lhs_bo.lower_bound,
                                    upper_bound=rhs_bo.lower_bound - 1)

                    return True, [ (lhs, r) ]
                elif lhs_bo.lower_bound >= rhs_bo.lower_bound:
                    r = self._backend.CreateStridedInterval(bits=rhs_bo.bits,
                                    stride=lhs_bo.stride,
                                    lower_bound=rhs_bo.lower_bound + 1,
                                    upper_bound=lhs_bo.upper_bound)

                    return True, [ (lhs, r) ]
                else:
                    # We cannot handle it precisely
                    return True, [ ]
        else:
            # TODO: handle this
            return True, [ ]

def is_true(a): return backends.vsa.is_true(a)
def is_false(a): return backends.vsa.is_false(a)

from .errors import ClaripyBalancerError, ClaripyBackendVSAError, BackendError
from .ast.base import Base
from . import _all_operations
from .backend_manager import backends
from . import vsa
from . import bv



import sys
import logging
import threading
l = logging.getLogger("claripy.backends.backend_z3")

if sys.platform == 'darwin':
    z3_library_file = "libz3.dylib"
elif sys.platform == 'win32':
    z3_library_file = "libz3.dll"
else:
    z3_library_file = "libz3.so"


solve_count = 0
cache_count = 0

from decimal import Decimal
import weakref

# import and set up Z3
import os
import z3

from ..errors import ClaripyZ3Error

_z3_paths = [ ]

if "Z3PATH" in os.environ:
    _z3_paths.append(os.environ["Z3PATH"])
if "VIRTUAL_ENV" in os.environ:
    virtual_env = os.environ["VIRTUAL_ENV"]
    _z3_paths.append(os.path.join(os.environ["VIRTUAL_ENV"], "lib"))
_z3_paths.extend(sys.path)
_z3_paths.append("/usr/lib")
_z3_paths.append("/usr/local/lib")
_z3_paths.append("/opt/python/lib")

for z3_path in _z3_paths:
    if not '.so' in z3_path and not '.dll' in z3_path:
        z3_path = os.path.join(z3_path, z3_library_file)
    if os.path.exists(z3_path):
        z3.init(z3_path)
        break
else:
    raise ClaripyZ3Error("Unable to find %s", z3_library_file)

supports_fp = hasattr(z3, 'fpEQ')

from ..backend import Backend

#pylint:disable=unidiomatic-typecheck

def condom(f):
    def z3_condom(*args, **kwargs):
        '''
        The Z3 condom intersects Z3Exceptions and throws a ClaripyZ3Error instead.
        '''
        try:
            condom_args = tuple((int(a) if type(a) is long and a < sys.maxint else a) for a in args)
            return f(*condom_args, **kwargs)
        except z3.Z3Exception as ze:
            _, _, traceback = sys.exc_info()
            raise ClaripyZ3Error, ("Z3Exception: %s" % ze), traceback
    return z3_condom

def _raw_caller(f):
    @staticmethod
    @condom
    def raw_caller(*args, **kwargs):
        return f(*args, **kwargs)
    return raw_caller

class BackendZ3(Backend):
    _split_on = { 'And', 'Or' }

    def __init__(self):
        Backend.__init__(self, solver_required=True)
        self._enable_simplification_cache = False

        # and the operations
        all_ops = backend_fp_operations | backend_operations if supports_fp else backend_operations
        for o in all_ops - {'BVV', 'BoolV', 'FPV', 'FP', 'BitVec'}:
            self._op_raw[o] = getattr(self, '_op_raw_' + o)

        self._op_raw['__ge__'] = self._op_raw_UGE
        self._op_raw['__gt__'] = self._op_raw_UGT
        self._op_raw['__le__'] = self._op_raw_ULE
        self._op_raw['__lt__'] = self._op_raw_ULT

        self._op_raw['Reverse'] = self._op_raw_Reverse
        self._op_raw['Identical'] = self._identical
        self._op_raw['fpToSBV'] = self._op_raw_fpToSBV
        self._op_raw['fpToUBV'] = self._op_raw_fpToUBV

        self._op_expr['BVS'] = self.BVS
        self._op_expr['BVV'] = self.BVV
        self._op_expr['FPV'] = self.FPV
        self._op_expr['FPS'] = self.FPS
        self._op_expr['BoolV'] = self.BoolV
        self._op_expr['BoolS'] = self.BoolS

    @property
    def _context(self):
        try:
            return self._tls.context
        except AttributeError:
            self._tls.context = z3.Context() if threading.current_thread().name != 'MainThread' else z3.main_ctx()
            return self._tls.context

    @property
    def _ast_cache(self):
        try:
            return self._tls.ast_cache
        except AttributeError:
            self._tls.ast_cache = weakref.WeakValueDictionary()
            return self._tls.ast_cache

    @property
    def _var_cache(self):
        try:
            return self._tls.var_cache
        except AttributeError:
            self._tls.var_cache = weakref.WeakValueDictionary()
            return self._tls.var_cache

    @property
    def _sym_cache(self):
        try:
            return self._tls.sym_cache
        except AttributeError:
            self._tls.sym_cache = weakref.WeakValueDictionary()
            return self._tls.sym_cache

    @property
    def _simplification_cache_key(self):
        try:
            return self._tls.simplification_cache_key
        except AttributeError:
            self._tls.simplification_cache_key = weakref.WeakValueDictionary()
            return self._tls.simplification_cache_key

    @property
    def _simplification_cache_val(self):
        try:
            return self._tls.simplification_cache_val
        except AttributeError:
            self._tls.simplification_cache_val = weakref.WeakValueDictionary()
            return self._tls.simplification_cache_val

    def downsize(self):
        Backend.downsize(self)

        self._ast_cache.clear()
        self._var_cache.clear()
        self._sym_cache.clear()
        self._simplification_cache_key.clear()
        self._simplification_cache_val.clear()

    @condom
    def _size(self, e, result=None):
        if not isinstance(e, z3.BitVecRef) and not isinstance(e, z3.BitVecNumRef):
            l.debug("Unable to determine length of value of type %s", e.__class__)
            raise BackendError("Unable to determine length of value of type %s" % e.__class__)
        return e.size()

    def _name(self, e, result=None): #pylint:disable=unused-argument
        l.warning("BackendZ3.name() called. This is weird.")
        raise BackendError("name is not implemented yet")

    #
    # Core creation methods
    #

    @condom
    def BVS(self, ast, result=None): #pylint:disable=unused-argument
        name, mn, mx, stride, _ = ast.args
        size = ast.size()
        expr = z3.BitVec(name, size, ctx=self._context)
        if mn is not None:
            expr = z3.If(z3.ULT(expr, mn), mn, expr, ctx=self._context)
        if mx is not None:
            expr = z3.If(z3.UGT(expr, mx), mx, expr, ctx=self._context)
        if stride is not None:
            expr = (expr / stride) * stride
        return expr

    @condom
    def BVV(self, ast, result=None): #pylint:disable=unused-argument
        if ast.args[0] is None:
            raise BackendError("Z3 can't handle empty BVVs")

        size = ast.size()
        return z3.BitVecVal(ast.args[0], size, ctx=self._context)

    @staticmethod
    @condom
    def FPS(name, sort): #pylint:disable=unused-argument
        raise BackendError("TODO: not sure how to do this")

    @condom
    def FPV(self, ast, result=None): #pylint:disable=unused-argument
        val = str(ast.args[0])
        sort = self._convert(ast.args[1])
        if val == 'inf':
            return z3.fpPlusInfinity(sort)
        elif val == '-inf':
            return z3.fpMinusInfinity(sort)
        elif val == '0.0':
            return z3.fpPlusZero(sort)
        elif val == '-0.0':
            return z3.fpMinusZero(sort)
        elif val == 'nan':
            return z3.fpNaN(sort)
        else:
            better_val = str(Decimal(ast.args[0]))
            return z3.FPVal(better_val, sort, ctx=self._context)

    @condom
    def BoolS(self, ast, result=None): #pylint:disable=unused-argument
        return z3.Bool(ast.args[0], ctx=self._context)

    @condom
    def BoolV(self, ast, result=None): #pylint:disable=unused-argument
        return z3.BoolVal(ast.args[0], ctx=self._context)

    #
    # Conversions
    #

    @condom
    def _convert(self, obj, result=None):
        if isinstance(obj, FSort):
            return z3.FPSort(obj.exp, obj.mantissa, ctx=self._context)
        elif isinstance(obj, RM):
            if obj == RM_RNE:
                return z3.RNE(ctx=self._context)
            elif obj == RM_RNA:
                return z3.RNA(ctx=self._context)
            elif obj == RM_RTP:
                return z3.RTP(ctx=self._context)
            elif obj == RM_RTN:
                return z3.RTN(ctx=self._context)
            elif obj == RM_RTZ:
                return z3.RTZ(ctx=self._context)
            else:
                raise BackendError("unrecognized rounding mode")
        elif obj is True:
            return z3.BoolVal(True, ctx=self._context)
        elif obj is False:
            return z3.BoolVal(False, ctx=self._context)
        elif type(obj) in (int, long, float, str):
            return obj
        elif hasattr(obj, '__module__') and obj.__module__ == 'z3':
            return obj
        else:
            l.debug("BackendZ3 encountered unexpected type %s", type(obj))
            raise BackendError("unexpected type %s encountered in BackendZ3" % type(obj))

    def call(self, *args, **kwargs):
        return Backend.call(self, *args, **kwargs)

    @condom
    def _abstract(self, z):
        #return self._abstract(z, split_on=split_on)[0]
        return self._abstract_internal(z.ctx.ctx, z.ast)

    @staticmethod
    def _z3_ast_hash(ctx, ast):
        """
        This is a better hashing function for z3 Ast objects. Z3_get_ast_hash() creates too many hash collisions.

        :param ctx: z3 Context
        :param ast: z3 Ast object
        :return: An integer - the hash
        """

        z3_hash = z3.Z3_get_ast_hash(ctx, ast)
        z3_ast_ref = ast.value # this seems to be the memory address
        z3_sort = z3.Z3_get_sort(ctx, ast).value
        return hash("%d_%d_%d" % (z3_hash, z3_sort, z3_ast_ref))

    def _abstract_internal(self, ctx, ast, split_on=None):
        h = self._z3_ast_hash(ctx, ast)
        try:
            return self._ast_cache[h]
        except KeyError:
            pass

        decl = z3.Z3_get_app_decl(ctx, ast)
        decl_num = z3.Z3_get_decl_kind(ctx, decl)
        z3_sort = z3.Z3_get_sort(ctx, ast)

        if decl_num not in z3_op_nums:
            raise ClaripyError("unknown decl kind %d" % decl_num)
        if z3_op_nums[decl_num] not in op_map:
            raise ClaripyError("unknown decl op %s" % z3_op_nums[decl_num])
        op_name = op_map[z3_op_nums[decl_num]]

        num_args = z3.Z3_get_app_num_args(ctx, ast)
        split_on = self._split_on if split_on is None else split_on
        new_split_on = split_on if op_name in split_on else set()
        children = [ self._abstract_internal(ctx, z3.Z3_get_app_arg(ctx, ast, i), new_split_on) for i in range(num_args) ]

        append_children = True

        if op_name == 'True':
            return BoolV(True)
        elif op_name == 'False':
            return BoolV(False)
        elif op_name.startswith('RM_'):
            return RM.from_name(op_name)
        elif op_name == 'BitVecVal':
            bv_num = long(z3.Z3_get_numeral_string(ctx, ast))
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
            return BVV(bv_num, bv_size)
        elif op_name == 'FPVal':
            # this is really imprecise
            fp_mantissa = float(z3.Z3_fpa_get_numeral_significand_string(ctx, ast))
            fp_exp = long(z3.Z3_fpa_get_numeral_exponent_string(ctx, ast))
            value = fp_mantissa * (2 ** fp_exp)

            ebits = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            sbits = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(ebits, sbits)

            return FPV(value, sort)
        elif op_name in ('MinusZero', 'MinusInf', 'PlusZero', 'PlusInf', 'NaN'):
            ebits = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            sbits = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(ebits, sbits)

            if op_name == 'MinusZero':
                return FPV(-0.0, sort)
            elif op_name == 'MinusInf':
                return FPV(float('-inf'), sort)
            elif op_name == 'PlusZero':
                return FPV(0.0, sort)
            elif op_name == 'PlusInf':
                return FPV(float('inf'), sort)
            elif op_name == 'NaN':
                return FPV(float('nan'), sort)
        elif op_name == 'UNINTERPRETED' and num_args == 0: # this *might* be a BitVec ;-)
            bv_name = z3.Z3_get_symbol_string(ctx, z3.Z3_get_decl_name(ctx, decl))
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)

            #if bv_name.count('_') < 2:
            #      import ipdb; ipdb.set_trace()
            return BV("BVS", (bv_name, None, None, None, False), length=bv_size, variables={ bv_name }, symbolic=True)
        elif op_name == 'UNINTERPRETED':
            mystery_name = z3.Z3_get_symbol_string(ctx, z3.Z3_get_decl_name(ctx, decl))
            args = [ ]
            if mystery_name == 'bvsmod_i':
                op_name = '__mod__'
                decl_num = z3.Z3_OP_BSMOD
            elif mystery_name == 'bvsdiv_i':
                op_name = '__div__'
                decl_num = z3.Z3_OP_BSDIV
            else:
                l.error("Mystery operation %s in BackendZ3._abstract_internal. Please report this.", mystery_name)
        elif op_name == 'Extract':
            hi = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            lo = z3.Z3_get_decl_int_parameter(ctx, decl, 1)
            args = [ hi, lo ]
        elif op_name in ('SignExt', 'ZeroExt'):
            num = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            args = [ num ]
        elif op_name in ('fpToFP', 'fpToFPSigned'):
            exp = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            mantissa = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(exp, mantissa)
            args = children + [sort]
            append_children = False
        elif op_name in ('fpToSBV', 'fpToUBV'):
            # uuuuuugggggghhhhhh
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
            args = children + [bv_size]
            append_children = False
        else:
            args = [ ]

        if append_children:
            args.extend(children)

        # fix up many-arg __add__
        if op_name in bin_ops and len(args) > 2:
            many_args = args #pylint:disable=unused-variable
            last = args[-1]
            rest = args[:-1]

            a = args[0].make_like(op_name, rest[:2])
            for b in rest[2:]:
                a = args[0].make_like(op_name, [a,b])
            args = [ a, last ]

        # hmm.... honestly not sure what to do here
        result_ty = op_type_map[z3_op_nums[decl_num]]
        ty = type(args[-1])

        if op_name == 'If':
            # If is polymorphic and thus must be handled specially
            ty = type(args[1])

            a = ty('If', tuple(args), length=args[1].length)
        else:
            if hasattr(ty, op_name) or hasattr(_all_operations, op_name):
                op = getattr(ty if hasattr(ty, op_name) else _all_operations, op_name)
                if op.calc_length is not None:
                    length = op.calc_length(*args)
                    a = result_ty(op_name, tuple(args), length=length)
                else:
                    a = result_ty(op_name, tuple(args))
            else:
                a = result_ty(op_name, tuple(args))

        self._ast_cache[h] = a
        return a

    @staticmethod
    def _abstract_to_primitive(ctx, ast):
        decl = z3.Z3_get_app_decl(ctx, ast)
        decl_num = z3.Z3_get_decl_kind(ctx, decl)

        if decl_num not in z3_op_nums:
            raise ClaripyError("unknown decl kind %d" % decl_num)
        if z3_op_nums[decl_num] not in op_map:
            raise ClaripyError("unknown decl op %s" % z3_op_nums[decl_num])
        op_name = op_map[z3_op_nums[decl_num]]

        if op_name == 'BitVecVal':
            return long(z3.Z3_get_numeral_string(ctx, ast))
        elif op_name == 'FPVal':
            # this is really imprecise
            fp_mantissa = float(z3.Z3_fpa_get_numeral_significand_string(ctx, ast))
            fp_exp = long(z3.Z3_fpa_get_numeral_exponent_string(ctx, ast))
            return fp_mantissa * (2 ** fp_exp)
        elif op_name == 'True':
            return True
        elif op_name == 'False':
            return False
        else:
            raise BackendError("Unable to abstract Z3 object to primitive")

    def solver(self, timeout=None):
        s = z3.Solver(ctx=self._context)
        if timeout is not None:
            if 'soft_timeout' in str(s.param_descrs()):
                s.set('soft_timeout', timeout)
                s.set('solver2_timeout', timeout)
            else:
                s.set('timeout', timeout)
        return s

    def _add(self, s, c):
        s.add(*c)

    def _check(self, s, extra_constraints=()):
        return self._check_and_model(s, extra_constraints=extra_constraints)[0]

    @condom
    def _check_and_model(self, s, extra_constraints=()): #pylint:disable=no-self-use
        global solve_count

        solve_count += 1
        if len(extra_constraints) > 0:
            s.push()
            s.add(*extra_constraints)

        l.debug("Doing a check!")
        #print "CHECKING"
        satness = s.check() == z3.sat
        if satness:
            model = s.model()
        else:
            model = None
        #print "CHECKED"

        if len(extra_constraints) > 0:
            s.pop()
        return satness, model

    def _primitive_from_model(self, model, expr):
        v = model.eval(expr, model_completion=True)
        return self._abstract_to_primitive(v.ctx.ctx, v.ast)

    @condom
    def _results(self, s, extra_constraints=(), generic_model=True):
        satness, z3_model = self._check_and_model(s, extra_constraints=extra_constraints)
        model = { }

        if satness:
            l.debug("sat!")
            if generic_model:
                for m_f in z3_model:
                    n = m_f.name()
                    m = m_f()
                    me = z3_model.eval(m)
                    model[n] = self._abstract_to_primitive(me.ctx.ctx, me.ast)
        else:
            l.debug("unsat!")

        return Result(satness, model, backend_model=z3_model)

    @condom
    def _eval(self, expr, n, extra_constraints=(), result=None, solver=None):
        r = self._batch_eval([ expr ], n, extra_constraints=extra_constraints, result=result, solver=solver)
        # Unpack it
        return [ x[0] for x in r ]

    @condom
    def _batch_eval(self, exprs, n, extra_constraints=(), result=None, solver=None):
        global solve_count, cache_count

        results = [ ]
        model = result.backend_model if result else None
        if len(extra_constraints) > 0 or n != 1:
            solver.push()
        if len(extra_constraints) > 0:
            solver.add(*extra_constraints)
            model = None
            l.debug("Disregarding cache")

        for i in range(n):
            if model is None:
                solve_count += 1
                l.debug("Doing a check!")
                if solver.check() == z3.sat:
                    model = solver.model()
            else:
                cache_count += 1

            if model is None:
                break

            # construct results
            r = [ ]
            for expr in exprs:
                if not type(expr) in {int, float, str, bool, long}:
                    v = self._primitive_from_model(model, expr)
                    r.append(v)
                else:
                    r.append(expr)

            # Append the solution to the result list
            results.append(tuple(r))

            # Construct the extra constraint so we don't get the same result anymore
            if i + 1 != n:
                solver.add(z3.Not(z3.And(*[(ex == ex_v) for ex, ex_v in zip(exprs, r)])))
                model = None

        if len(extra_constraints) > 0 or n != 1:
            solver.pop()

        if len(results) == 0:
            raise UnsatError("constraints are unsat")

        return results

    def _max(self, expr, extra_constraints=(), result=None, solver=None):
        return max(self._max_values(expr, extra_constraints=extra_constraints, result=result, solver=solver))

    def _min(self, expr, extra_constraints=(), result=None, solver=None):
        return min(self._min_values(expr, extra_constraints=extra_constraints, result=result, solver=solver))

    @condom
    def _min_values(self, expr, extra_constraints=(), result=None, solver=None):
        global solve_count

        lo = 0
        hi = 2**expr.size()-1
        vals = set()

        numpop = 0
        if len(extra_constraints) > 0:
            solver.push()
            numpop += 1
            solver.add(*[self.convert(e) for e in extra_constraints])

        while hi-lo > 1:
            middle = (lo + hi)/2
            #l.debug("h/m/l/d: %d %d %d %d", hi, middle, lo, hi-lo)

            solver.push()
            solver.add(z3.UGE(expr, lo), z3.ULE(expr, middle))
            numpop += 1

            solve_count += 1
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                l.debug("... still sat")
                vals.add(self._primitive_from_model(solver.model(), expr))
                hi = middle
            else:
                l.debug("... now unsat")
                lo = middle
                solver.pop()
                numpop -= 1

        for _ in range(numpop):
            solver.pop()

        #l.debug("final hi/lo: %d, %d", hi, lo)

        if hi == lo: return lo
        else:
            solver.push()
            solver.add(expr == lo)
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                vals.add(lo)
                solver.pop()
            else:
                vals.add(hi)
                solver.pop()

        return vals

    @condom
    def _max_values(self, expr, extra_constraints=(), result=None, solver=None):
        global solve_count

        lo = 0
        hi = 2**expr.size()-1
        vals = set()

        numpop = 0
        if len(extra_constraints) > 0:
            solver.push()
            numpop += 1
            solver.add(*[self.convert(e) for e in extra_constraints])

        while hi-lo > 1:
            middle = (lo + hi)/2
            #l.debug("h/m/l/d: %d %d %d %d", hi, middle, lo, hi-lo)

            solver.push()
            solver.add(z3.UGT(expr, middle), z3.ULE(expr, hi))
            numpop += 1

            solve_count += 1
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                l.debug("... still sat")
                lo = middle
                vals.add(self._primitive_from_model(solver.model(), expr))
            else:
                l.debug("... now unsat")
                hi = middle
                solver.pop()
                numpop -= 1
            #l.debug("        now: %d %d %d %d", hi, middle, lo, hi-lo)

        for _ in range(numpop):
            solver.pop()

        if hi == lo:
            vals.add(hi)
        else:
            solver.push()
            solver.add(expr == hi)
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                vals.add(hi)
                solver.pop()
            else:
                vals.add(lo)
                solver.pop()

        return vals

    def _simplify(self, expr): #pylint:disable=W0613,R0201
        raise Exception("This shouldn't be called. Bug Yan.")

    @condom
    def simplify(self, expr):
        if expr._simplified:
            return expr

        if self._enable_simplification_cache:
            try:
                k = self._simplification_cache_key[expr._cache_key]
                #print "HIT WEAK KEY CACHE"
                return k
            except KeyError:
                pass
            try:
                k = self._simplification_cache_val[expr._cache_key]
                #print "HIT WEAK VALUE CACHE"
                return k
            except KeyError:
                pass

            #print "MISS CACHE"

        l.debug("SIMPLIFYING EXPRESSION")

        #print "SIMPLIFYING"

        expr_raw = self.convert(expr)

        #l.debug("... before: %s (%s)", expr_raw, expr_raw.__class__.__name__)

        #s = expr_raw
        if isinstance(expr_raw, z3.BoolRef):
            tactics = z3.Then(
                z3.Tactic("simplify", ctx=self._context),
                z3.Tactic("propagate-ineqs", ctx=self._context),
                z3.Tactic("propagate-values", ctx=self._context),
                z3.Tactic("unit-subsume-simplify", ctx=self._context),
                ctx=self._context
            )
            s = tactics(expr_raw).as_expr()
            #n = s.decl().name()
            #if n == 'true':
            #   s = True
            #elif n == 'false':
            #   s = False
        elif isinstance(expr_raw, z3.BitVecRef):
            s = z3.simplify(expr_raw)
        else:
            s = expr_raw

        o = self._abstract(s)
        o._simplified = Base.FULL_SIMPLIFY

        if self._enable_simplification_cache:
            self._simplification_cache_val[expr._cache_key] = o
            self._simplification_cache_key[expr._cache_key] = o
        return o

    def _is_false(self, e, extra_constraints=(), result=None, solver=None):
        return z3.simplify(e).eq(z3.BoolVal(False, ctx=self._context))

    def _is_true(self, e, extra_constraints=(), result=None, solver=None):
        return z3.simplify(e).eq(z3.BoolVal(True, ctx=self._context))

    def _solution(self, expr, v, result=None, extra_constraints=(), solver=None):
        return self._check(solver, extra_constraints=(expr == v,) + tuple(extra_constraints))

    #
    # Some Z3 passthroughs
    #

    # these require the context

    def _op_raw_And(self, *args):
        return z3.And(*(tuple(args) + ( self._context, )))

    def _op_raw_Or(self, *args):
        return z3.Or(*(tuple(args) + ( self._context, )))

    def _op_raw_Not(self, a):
        return z3.Not(a, ctx=self._context)

    def _op_raw_If(self, i, t, e):
        return z3.If(i, t, e, ctx=self._context)

    @condom
    def _op_raw_fpToSBV(self, rm, fp, bv_len):
        return z3.fpToSBV(rm, fp, z3.BitVecSort(bv_len, ctx=self._context))

    @condom
    def _op_raw_fpToUBV(self, rm, fp, bv_len):
        return z3.fpToUBV(rm, fp, z3.BitVecSort(bv_len, ctx=self._context))

    # and these do not
    _op_raw_Concat = _raw_caller(z3.Concat)
    _op_raw_Extract = _raw_caller(z3.Extract)
    _op_raw_LShR = _raw_caller(z3.LShR)
    _op_raw_RotateLeft = _raw_caller(z3.RotateLeft)
    _op_raw_RotateRight = _raw_caller(z3.RotateRight)
    _op_raw_SignExt = _raw_caller(z3.SignExt)
    _op_raw_UGE = _raw_caller(z3.UGE)
    _op_raw_UGT = _raw_caller(z3.UGT)
    _op_raw_ULE = _raw_caller(z3.ULE)
    _op_raw_ULT = _raw_caller(z3.ULT)
    _op_raw_ZeroExt = _raw_caller(z3.ZeroExt)
    _op_raw_fpAbs = _raw_caller(z3.fpAbs)
    _op_raw_fpAdd = _raw_caller(z3.fpAdd)
    _op_raw_fpDiv = _raw_caller(z3.fpDiv)
    _op_raw_fpEQ = _raw_caller(z3.fpEQ)
    _op_raw_fpFP = _raw_caller(z3.fpFP)
    _op_raw_fpGEQ = _raw_caller(z3.fpGEQ)
    _op_raw_fpGT = _raw_caller(z3.fpGT)
    _op_raw_fpLEQ = _raw_caller(z3.fpLEQ)
    _op_raw_fpLT = _raw_caller(z3.fpLT)
    _op_raw_fpMul = _raw_caller(z3.fpMul)
    _op_raw_fpNeg = _raw_caller(z3.fpNeg)
    _op_raw_fpSub = _raw_caller(z3.fpSub)
    _op_raw_fpToFP = _raw_caller(z3.fpToFP)
    _op_raw_fpToIEEEBV = _raw_caller(z3.fpToIEEEBV)

    @staticmethod
    @condom
    def _op_raw_Reverse(a):
        if a.size() == 8:
            return a
        elif a.size() % 8 != 0:
            raise ClaripyOperationError("can't reverse non-byte sized bitvectors")
        else:
            return z3.Concat(*[z3.Extract(i+7, i, a) for i in range(0, a.size(), 8)])

    @staticmethod
    @condom
    def _op_raw_SLT(a, b):
        return a < b

    @staticmethod
    @condom
    def _op_raw_SLE(a, b):
        return a <= b

    @staticmethod
    @condom
    def _op_raw_SGT(a, b):
        return a > b

    @staticmethod
    @condom
    def _op_raw_SGE(a, b):
        return a >= b

    def _identical(self, a, b, result=None):
        return a.eq(b)

#
# this is for the actual->abstract conversion above
#

# these are Z3 operation names for abstraction
z3_op_names = [ _ for _ in dir(z3) if _.startswith('Z3_OP') ]
z3_op_nums = { getattr(z3, _): _ for _ in z3_op_names }

#pylint:disable=bad-continuation
op_map = {
    # Boolean
    'Z3_OP_TRUE': 'True',
    'Z3_OP_FALSE': 'False',
    'Z3_OP_EQ': '__eq__',
    'Z3_OP_DISTINCT': '__ne__',
    'Z3_OP_ITE': 'If',
    'Z3_OP_AND': 'And',
    'Z3_OP_OR': 'Or',
    'Z3_OP_IFF': '__eq__',
    'Z3_OP_XOR': 'Xor',
    'Z3_OP_NOT': 'Not',
    'Z3_OP_IMPLIES': 'Implies',
    #'Z3_OP_OEQ': None,

    # Arithmetic
    #'Z3_OP_ANUM': None,
    #'Z3_OP_AGNUM': None,
    'Z3_OP_LE': 'SLE',
    'Z3_OP_GE': 'SGE',
    'Z3_OP_LT': 'SLT',
    'Z3_OP_GT': 'SGT',
    'Z3_OP_ADD': '__add__',
    'Z3_OP_SUB': '__sub__',
    'Z3_OP_UMINUS': '__neg__',
    'Z3_OP_MUL': '__mul__',
    'Z3_OP_DIV': '__div__',
    'Z3_OP_IDIV': '__div__',
    'Z3_OP_REM': '__mod__', # TODO: is this correct?
    'Z3_OP_MOD': '__mod__',
    #'Z3_OP_TO_REAL': None,
    #'Z3_OP_TO_INT': None,
    #'Z3_OP_IS_INT': None,
    'Z3_OP_POWER': '__pow__',

    # Arrays & Sets
    #'Z3_OP_STORE': None,
    #'Z3_OP_SELECT': None,
    #'Z3_OP_CONST_ARRAY': None,
    #'Z3_OP_ARRAY_MAP': None,
    #'Z3_OP_ARRAY_DEFAULT': None,
    #'Z3_OP_SET_UNION': None,
    #'Z3_OP_SET_INTERSECT': None,
    #'Z3_OP_SET_DIFFERENCE': None,
    #'Z3_OP_SET_COMPLEMENT': None,
    #'Z3_OP_SET_SUBSET': None,
    #'Z3_OP_AS_ARRAY': None,

    # Bit-vectors
    'Z3_OP_BNUM': 'BitVecVal',
    #'Z3_OP_BIT1': None, # MAYBE TODO
    #'Z3_OP_BIT0': None, # MAYBE TODO
    'Z3_OP_BNEG': '__neg__',
    'Z3_OP_BADD': '__add__',
    'Z3_OP_BSUB': '__sub__',
    'Z3_OP_BMUL': '__mul__',

    'Z3_OP_BSDIV': '__div__',
    'Z3_OP_BUDIV': '__div__', # TODO: is this correct?
    'Z3_OP_BSREM': '__mod__', # TODO: is this correct?
    'Z3_OP_BUREM': '__mod__', # TODO: is this correct?
    'Z3_OP_BSMOD': '__mod__', # TODO: is this correct?

    # special functions to record the division by 0 cases
    # these are internal functions
    #'Z3_OP_BSDIV0': None,
    #'Z3_OP_BUDIV0': None,
    #'Z3_OP_BSREM0': None,
    #'Z3_OP_BUREM0': None,
    #'Z3_OP_BSMOD0': None,

    'Z3_OP_ULEQ': 'ULE',
    'Z3_OP_SLEQ': 'SLE',
    'Z3_OP_UGEQ': 'UGE',
    'Z3_OP_SGEQ': 'SGE',
    'Z3_OP_ULT': 'ULT',
    'Z3_OP_SLT': 'SLT',
    'Z3_OP_UGT': 'UGT',
    'Z3_OP_SGT': 'SGT',

    'Z3_OP_BAND': '__and__',
    'Z3_OP_BOR': '__or__',
    'Z3_OP_BNOT': '__invert__',
    'Z3_OP_BXOR': '__xor__',
    #'Z3_OP_BNAND': None,
    #'Z3_OP_BNOR': None,
    #'Z3_OP_BXNOR': None,

    'Z3_OP_CONCAT': 'Concat',
    'Z3_OP_SIGN_EXT': 'SignExt',
    'Z3_OP_ZERO_EXT': 'ZeroExt',
    'Z3_OP_EXTRACT': 'Extract',
    'Z3_OP_REPEAT': 'RepeatBitVec',

    #'Z3_OP_BREDOR': None,
    #'Z3_OP_BREDAND': None,
    #'Z3_OP_BCOMP': None,

    'Z3_OP_BSHL': '__lshift__',
    'Z3_OP_BLSHR': 'LShR',
    'Z3_OP_BASHR': '__rshift__',
    #'Z3_OP_ROTATE_LEFT': None,
    #'Z3_OP_ROTATE_RIGHT': None,
    'Z3_OP_EXT_ROTATE_LEFT': 'RotateLeft',
    'Z3_OP_EXT_ROTATE_RIGHT': 'RotateRight',

    'Z3_OP_FPA_TO_SBV': 'fpToSBV',
    'Z3_OP_FPA_TO_UBV': 'fpToUBV',
    'Z3_OP_FPA_TO_IEEE_BV': 'fpToIEEEBV',
    'Z3_OP_FPA_TO_FP': 'fpToFP',
    'Z3_OP_FPA_NUM': 'FPVal',

    'Z3_OP_FPA_MINUS_ZERO': 'MinusZero',
    'Z3_OP_FPA_MINUS_INF': 'MinusInf',
    'Z3_OP_FPA_PLUS_ZERO': 'PlusZero',
    'Z3_OP_FPA_PLUS_INF': 'PlusInf',
    'Z3_OP_FPA_NAN': 'NaN',

    'Z3_OP_FPA_EQ': 'fpEQ',
    'Z3_OP_FPA_GT': 'fpGT',
    'Z3_OP_FPA_GE': 'fpGEQ',
    'Z3_OP_FPA_LT': 'fpLT',
    'Z3_OP_FPA_LE': 'fpLEQ',

    'Z3_OP_FPA_ABS': 'fpAbs',
    'Z3_OP_FPA_NEG': 'fpNeg',
    'Z3_OP_FPA_ADD': 'fpAdd',
    'Z3_OP_FPA_SUB': 'fpSub',
    'Z3_OP_FPA_MUL': 'fpMul',
    'Z3_OP_FPA_DIV': 'fpDiv',

    'Z3_OP_FPA_RM_NEAREST_TIES_TO_EVEN': 'RM_RNE',
    'Z3_OP_FPA_RM_NEAREST_TIES_TO_AWAY': 'RM_RNA',
    'Z3_OP_FPA_RM_TOWARD_ZERO': 'RM_RTZ',
    'Z3_OP_FPA_RM_TOWARD_POSITIVE': 'RM_RTP',
    'Z3_OP_FPA_RM_TOWARD_NEGATIVE': 'RM_RTN',

    'Z3_OP_UNINTERPRETED': 'UNINTERPRETED'
}

from ..ast.base import Base
from ..ast.bv import BV, BVV
from ..ast.bool import BoolV, Bool
from ..ast.fp import FP, FPV
from ..operations import backend_operations, backend_fp_operations, bin_ops
from ..result import Result
from ..fp import FSort, RM, RM_RNE, RM_RNA, RM_RTP, RM_RTN, RM_RTZ
from ..errors import ClaripyError, BackendError, UnsatError, ClaripyOperationError
from .. import _all_operations

op_type_map = {
    # Boolean
    'Z3_OP_TRUE': Bool,
    'Z3_OP_FALSE': Bool,
    'Z3_OP_EQ': Bool,
    'Z3_OP_DISTINCT': Bool,
    'Z3_OP_ITE': Bool,
    'Z3_OP_AND': Bool,
    'Z3_OP_OR': Bool,
    'Z3_OP_IFF': Bool,
    'Z3_OP_XOR': Bool,
    'Z3_OP_NOT': Bool,
    'Z3_OP_IMPLIES': Bool,
    #'Z3_OP_OEQ': None,

    # Arithmetic
    #'Z3_OP_ANUM': None,
    #'Z3_OP_AGNUM': None,
    'Z3_OP_LE': None,
    'Z3_OP_GE': None,
    'Z3_OP_LT': None,
    'Z3_OP_GT': None,
    'Z3_OP_ADD': None,
    'Z3_OP_SUB': None,
    'Z3_OP_UMINUS': None,
    'Z3_OP_MUL': None,
    'Z3_OP_DIV': None,
    'Z3_OP_IDIV': None,
    'Z3_OP_REM': None, # TODO: is this correct?
    'Z3_OP_MOD': None,
    #'Z3_OP_TO_REAL': None,
    #'Z3_OP_TO_INT': None,
    #'Z3_OP_IS_INT': None,
    'Z3_OP_POWER': None,

    # Arrays & Sets
    #'Z3_OP_STORE': None,
    #'Z3_OP_SELECT': None,
    #'Z3_OP_CONST_ARRAY': None,
    #'Z3_OP_ARRAY_MAP': None,
    #'Z3_OP_ARRAY_DEFAULT': None,
    #'Z3_OP_SET_UNION': None,
    #'Z3_OP_SET_INTERSECT': None,
    #'Z3_OP_SET_DIFFERENCE': None,
    #'Z3_OP_SET_COMPLEMENT': None,
    #'Z3_OP_SET_SUBSET': None,
    #'Z3_OP_AS_ARRAY': None,

    # Bit-vectors
    'Z3_OP_BNUM': 'BitVecVal',
    #'Z3_OP_BIT1': None, # MAYBE TODO
    #'Z3_OP_BIT0': None, # MAYBE TODO
    'Z3_OP_BNEG': BV,
    'Z3_OP_BADD': BV,
    'Z3_OP_BSUB': BV,
    'Z3_OP_BMUL': BV,

    'Z3_OP_BSDIV': BV,
    'Z3_OP_BUDIV': BV,
    'Z3_OP_BSREM': BV,
    'Z3_OP_BUREM': BV,
    'Z3_OP_BSMOD': BV,

    # special functions to record the division by 0 cases
    # these are internal functions
    #'Z3_OP_BSDIV0': None,
    #'Z3_OP_BUDIV0': None,
    #'Z3_OP_BSREM0': None,
    #'Z3_OP_BUREM0': None,
    #'Z3_OP_BSMOD0': None,

    'Z3_OP_ULEQ': Bool,
    'Z3_OP_SLEQ': Bool,
    'Z3_OP_UGEQ': Bool,
    'Z3_OP_SGEQ': Bool,
    'Z3_OP_ULT': Bool,
    'Z3_OP_SLT': Bool,
    'Z3_OP_UGT': Bool,
    'Z3_OP_SGT': Bool,

    'Z3_OP_BAND': BV,
    'Z3_OP_BOR': BV,
    'Z3_OP_BNOT': BV,
    'Z3_OP_BXOR': BV,
    #'Z3_OP_BNAND': None,
    #'Z3_OP_BNOR': None,
    #'Z3_OP_BXNOR': None,

    'Z3_OP_CONCAT': BV,
    'Z3_OP_SIGN_EXT': BV,
    'Z3_OP_ZERO_EXT': BV,
    'Z3_OP_EXTRACT': BV,
    'Z3_OP_REPEAT': BV,

    #'Z3_OP_BREDOR': None,
    #'Z3_OP_BREDAND': None,
    #'Z3_OP_BCOMP': None,

    'Z3_OP_BSHL': BV,
    'Z3_OP_BLSHR': BV,
    'Z3_OP_BASHR': BV,
    #'Z3_OP_ROTATE_LEFT': None,
    #'Z3_OP_ROTATE_RIGHT': None,
    'Z3_OP_EXT_ROTATE_LEFT': BV,
    'Z3_OP_EXT_ROTATE_RIGHT': BV,

    'Z3_OP_FPA_TO_SBV': BV,
    'Z3_OP_FPA_TO_UBV': BV,
    'Z3_OP_FPA_TO_IEEE_BV': BV,
    'Z3_OP_FPA_TO_FP': FP,
    'Z3_OP_FPA_NUM': FP,

    'Z3_OP_FPA_MINUS_ZERO': FP,
    'Z3_OP_FPA_MINUS_INF': FP,
    'Z3_OP_FPA_PLUS_ZERO': FP,
    'Z3_OP_FPA_PLUS_INF': FP,
    'Z3_OP_FPA_NAN': FP,

    'Z3_OP_FPA_EQ': Bool,
    'Z3_OP_FPA_GT': Bool,
    'Z3_OP_FPA_GE': Bool,
    'Z3_OP_FPA_LT': Bool,
    'Z3_OP_FPA_LE': Bool,

    'Z3_OP_FPA_ABS': FP,
    'Z3_OP_FPA_NEG': FP,
    'Z3_OP_FPA_ADD': FP,
    'Z3_OP_FPA_SUB': FP,
    'Z3_OP_FPA_MUL': FP,
    'Z3_OP_FPA_DIV': FP,

    'Z3_OP_UNINTERPRETED': 'UNINTERPRETED'
}



import functools
from .errors import ClaripyOperationError
from .backend_object import BackendObject

def compare_bits(f):
    @functools.wraps(f)
    def compare_guard(self, o):
        if self.bits != o.bits:
            raise TypeError("bitvectors are differently-sized (%d and %d)" % (self.bits, o.bits))
        return f(self, o)

    return compare_guard

def normalize_types(f):
    @functools.wraps(f)
    def normalize_helper(self, o):
        if hasattr(o, '__module__') and o.__module__ == 'z3':
            raise ValueError("this should no longer happen")
        if type(o) in (int, long):
            o = BVV(o, self.bits)
        if type(self) in (int, long):
            self = BVV(self, self.bits)

        if not isinstance(self, BVV) or not isinstance(o, BVV):
            return NotImplemented
        return f(self, o)

    return normalize_helper

class BVV(BackendObject):
    __slots__ = [ 'bits', '_value', 'mod', 'value' ]

    def __init__(self, value, bits):
        if bits == 0 or type(bits) not in (int, long) or type(value) not in (int, long):
            raise ClaripyOperationError("BVV needs a non-zero length and an int/long value")

        self.bits = bits
        self._value = 0
        self.mod = 2**bits
        self.value = value

    def __hash__(self):
        return hash((str(self.value), self.bits))

    def __getstate__(self):
        return (self.bits, self.value)

    def __setstate__(self, s):
        self.bits = s[0]
        self.mod = 2**self.bits
        self.value = s[1]

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._value = v % self.mod

    @property
    def signed(self):
        return self._value if self._value < self.mod/2 else self._value % (self.mod/2) - (self.mod/2)

    @signed.setter
    def signed(self, v):
        self._value = v % -self.mod

    #
    # Arithmetic stuff
    #

    @normalize_types
    @compare_bits
    def __add__(self, o):
        return BVV(self.value + o.value, self.bits)

    @normalize_types
    @compare_bits
    def __sub__(self, o):
        return BVV(self.value - o.value, self.bits)

    @normalize_types
    @compare_bits
    def __mul__(self, o):
        return BVV(self.value * o.value, self.bits)

    @normalize_types
    @compare_bits
    def __mod__(self, o):
        return BVV(self.value % o.value, self.bits)

    @normalize_types
    @compare_bits
    def __div__(self, o):
        return BVV(self.value / o.value, self.bits)

    #
    # Reverse arithmetic stuff
    #

    @normalize_types
    @compare_bits
    def __radd__(self, o):
        return BVV(self.value + o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rsub__(self, o):
        return BVV(o.value - self.value, self.bits)

    @normalize_types
    @compare_bits
    def __rmul__(self, o):
        return BVV(self.value * o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rmod__(self, o):
        return BVV(o.value % self.value, self.bits)

    @normalize_types
    @compare_bits
    def __rdiv__(self, o):
        return BVV(o.value / self.value, self.bits)

    #
    # Bit operations
    #

    @normalize_types
    @compare_bits
    def __and__(self, o):
        return BVV(self.value & o.value, self.bits)

    @normalize_types
    @compare_bits
    def __or__(self, o):
        return BVV(self.value | o.value, self.bits)

    @normalize_types
    @compare_bits
    def __xor__(self, o):
        return BVV(self.value ^ o.value, self.bits)

    @normalize_types
    @compare_bits
    def __lshift__(self, o):
        return BVV(self.value << o.signed, self.bits)

    @normalize_types
    @compare_bits
    def __rshift__(self, o):
        return BVV(self.signed >> o.signed, self.bits)

    def __invert__(self):
        return BVV(self.value ^ self.mod-1, self.bits)

    #
    # Reverse bit operations
    #

    @normalize_types
    @compare_bits
    def __rand__(self, o):
        return BVV(self.value & o.value, self.bits)

    @normalize_types
    @compare_bits
    def __ror__(self, o):
        return BVV(self.value | o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rxor__(self, o):
        return BVV(self.value ^ o.value, self.bits)

    @normalize_types
    @compare_bits
    def __rlshift__(self, o):
        return BVV(o.value << self.signed, self.bits)

    @normalize_types
    @compare_bits
    def __rrshift__(self, o):
        return BVV(o.signed >> self.signed, self.bits)

    #
    # Boolean stuff
    #

    @normalize_types
    @compare_bits
    def __eq__(self, o):
        return self.value == o.value

    @normalize_types
    @compare_bits
    def __ne__(self, o):
        return self.value != o.value

    @normalize_types
    @compare_bits
    def __lt__(self, o):
        return self.value < o.value

    @normalize_types
    @compare_bits
    def __gt__(self, o):
        return self.value > o.value

    @normalize_types
    @compare_bits
    def __le__(self, o):
        return self.value <= o.value

    @normalize_types
    @compare_bits
    def __ge__(self, o):
        return self.value >= o.value

    #
    # Conversions
    #

    def size(self):
        return self.bits

    def __repr__(self):
        return 'BVV(0x%x, %d)' % (self.value, self.bits)

#
# External stuff
#

def BitVecVal(value, bits):
    return BVV(value, bits)

def ZeroExt(num, o):
    return BVV(o.value, o.bits + num)

def SignExt(num, o):
    return BVV(o.signed, o.bits + num)

def Extract(f, t, o):
    return BVV((o.value >> t) & (2**(f+1) - 1), f-t+1)

def Concat(*args):
    total_bits = 0
    total_value = 0

    for o in args:
        total_value = (total_value << o.bits) | o.value
        total_bits += o.bits
    return BVV(total_value, total_bits)

def RotateRight(self, bits):
    return LShR(self, bits) | (self << (self.size()-bits))

def RotateLeft(self, bits):
    return (self << bits) | (LShR(self, (self.size()-bits)))

def Reverse(a):
    if a.size() == 8:
        return a
    elif a.size() % 8 != 0:
        raise ClaripyOperationError("can't reverse non-byte sized bitvectors")
    else:
        return Concat(*[Extract(i+7, i, a) for i in range(0, a.size(), 8)])


@normalize_types
@compare_bits
def ULT(self, o):
    return self.value < o.value

@normalize_types
@compare_bits
def UGT(self, o):
    return self.value > o.value

@normalize_types
@compare_bits
def ULE(self, o):
    return self.value <= o.value

@normalize_types
@compare_bits
def UGE(self, o):
    return self.value >= o.value

@normalize_types
@compare_bits
def SLT(self, o):
    return self.signed < o.signed

@normalize_types
@compare_bits
def SGT(self, o):
    return self.signed > o.signed

@normalize_types
@compare_bits
def SLE(self, o):
    return self.signed <= o.signed

@normalize_types
@compare_bits
def SGE(self, o):
    return self.signed >= o.signed


#
# Pure boolean stuff
#

def BoolV(b):
    return b

def And(*args):
    return all(args)

def Or(*args):
    return any(args)

def Not(b):
    return not b

@normalize_types
def normalizer(*args):
    return args

def If(c, t, f):
    t,f = normalizer(t,f) #pylint:disable=unbalanced-tuple-unpacking
    if c: return t
    else: return f

@normalize_types
@compare_bits
def LShR(a, b):
    return BVV(a.value >> b.signed, a.bits)


def test():
    a = BVV(1, 8)
    b = BVV(2, 8)
    assert a | b == 3
    assert a & b == 0
    assert a / b == 0
    assert b * b == 4
    assert a.signed == a.value
    assert a + 8 == 9

    c = BVV(128, 8)
    assert c.signed == -128

    d = BVV(255, 8)
    assert Extract(1, 0, d) == 3
    assert SignExt(8, d).value == 2**16-1
    assert ZeroExt(8, d).size() == 16
    assert ZeroExt(8, d).value == 255

    e = BVV(0b1010, 4)
    f = BVV(0b11, 2)
    assert Concat(e, e, e, e) == 0b1010101010101010
    assert Concat(e,f,f) == 0b10101111

if __name__ == '__main__':
    test()



import logging
l = logging.getLogger("claripy.backends.backend_concrete")

from ..backend import BackendError, Backend

class BackendConcrete(Backend):
    def __init__(self):
        Backend.__init__(self)
        self._make_raw_ops(set(backend_operations) - { 'If' }, op_module=bv)
        self._make_raw_ops(backend_fp_operations, op_module=fp)
        self._op_raw['If'] = self._If
        self._op_expr['BVS'] = self.BVS
        self._op_raw['BVV'] = self.BVV
        self._op_raw['FPV'] = self.FPV

    @staticmethod
    def BVV(value, size):
        if value is None:
            raise BackendError("can't handle empty BVVs")
        return bv.BVV(value, size)

    @staticmethod
    def FPV(op, sort):
        return fp.FPV(op, sort)

    @staticmethod
    def BVS(ast, result=None):
        name, mn, mx, stride, uninitialized = ast.args #pylint:disable=unused-variable

        if mn is not None and mn == mx:
            return bv.BVV(mx, ast.length)

        if result is None:
            l.debug("BackendConcrete can only handle BitVec when we are given a model")
            raise BackendError("BackendConcrete can only handle BitVec when we are given a model")
        if name not in result.model:
            l.debug("BackendConcrete needs variable %s in the model", name)
            raise BackendError("BackendConcrete needs variable %s in the model" % name)
        else:
            return bv.BVV(result.model[name], ast.length)

    def _If(self, b, t, f, result=None): #pylint:disable=no-self-use,unused-argument
        if not isinstance(b, bool):
            raise BackendError("BackendConcrete can't handle non-bool condition in If.")
        else:
            return t if b else f

    def _size(self, e, result=None):
        if type(e) in { bool, long, int }:
            return None
        elif type(e) in { bv.BVV }:
            return e.size()
        elif isinstance(e, fp.FPV):
            return e.sort.length
        else:
            raise BackendError("can't get size of type %s" % type(e))

    def _name(self, e, result=None): #pylint:disable=unused-argument,no-self-use
        return None

    def _identical(self, a, b, result=None):
        if type(a) is bv.BVV and type(b) is bv.BVV and a.size() != b.size():
            return False
        else:
            return a == b

    def _convert(self, a, result=None):
        if type(a) in { int, long, float, bool, str, bv.BVV, fp.FPV, fp.RM, fp.FSort }:
            return a
        raise BackendError("can't handle AST of type %s" % type(a))

    def _simplify(self, e):
        return e

    def _abstract(self, e): #pylint:disable=no-self-use
        if isinstance(e, bv.BVV):
            return BVV(e.value, e.size())
        elif isinstance(e, bool):
            return BoolV(e)
        elif isinstance(e, fp.FPV):
            return FPV(e.value, e.sort)
        else:
            raise BackendError("Couldn't abstract object of type {}".format(type(e)))

    def _cardinality(self, b, result=None):
        # if we got here, it's a cardinality of 1
        return 1

    #
    # Evaluation functions
    #

    @staticmethod
    def _to_primitive(expr):
        if type(expr) is bv.BVV:
            return expr.value
        elif type(expr) is fp.FPV:
            return expr.value
        elif type(expr) is bool:
            return expr
        elif type(expr) in (int, long):
            return expr

    def _eval(self, expr, n, result=None, solver=None, extra_constraints=()):
        if not all(extra_constraints):
            raise UnsatError('concrete False constraint in extra_constraints')

        return (self._to_primitive(expr),)

    def _batch_eval(self, exprs, n, result=None, extra_constraints=(), solver=None):
        if not all(extra_constraints):
            raise UnsatError('concrete False constraint in extra_constraints')

        return [ tuple([self._to_primitive(ex) for ex in exprs]) ]

    def _max(self, expr, result=None, solver=None, extra_constraints=()):
        if not all(extra_constraints):
            raise UnsatError('concrete False constraint in extra_constraints')
        return self._to_primitive(expr)

    def _min(self, expr, result=None, solver=None, extra_constraints=()):
        if not all(extra_constraints):
            raise UnsatError('concrete False constraint in extra_constraints')
        return self._to_primitive(expr)

    def _solution(self, expr, v, result=None, solver=None, extra_constraints=()):
        if not all(extra_constraints):
            raise UnsatError('concrete False constraint in extra_constraints')
        return self.convert(expr, result=result) == v

    def _is_true(self, e, extra_constraints=(), result=None, solver=None):
        return e == True
    def _is_false(self, e, extra_constraints=(), result=None, solver=None):
        return e == False
    def _has_true(self, e, extra_constraints=(), result=None, solver=None):
        return e == True
    def _has_false(self, e, extra_constraints=(), result=None, solver=None):
        return e == False

from ..operations import backend_operations, backend_fp_operations
from .. import bv, fp
from ..ast.bv import BVV
from ..ast.fp import FPV
from ..ast.bool import BoolV
from ..errors import UnsatError



import os
import sys
import struct
import weakref
import hashlib
import itertools
import cPickle as pickle

import logging
l = logging.getLogger("claripy.ast")

import ana

if os.environ.get('WORKER', False):
    WORKER = True
else:
    WORKER = False

md5_unpacker = struct.Struct('2Q')

#pylint:enable=unused-argument
#pylint:disable=unidiomatic-typecheck

def _inner_repr(a, **kwargs):
    if isinstance(a, Base):
        return a.__repr__(inner=True, **kwargs)
    else:
        return repr(a)

class ASTCacheKey(object):
    def __init__(self, a):
        self.ast = a

#
# AST variable naming
#

var_counter = itertools.count()
_unique_names = True

def _make_name(name, size, explicit_name=False, prefix=""):
    if _unique_names and not explicit_name:
        return "%s%s_%d_%d" % (prefix, name, var_counter.next(), size)
    else:
        return name


class Base(ana.Storable):
    '''
    An AST tracks a tree of operations on arguments. It has the following methods:

        op: the operation that is being done on the arguments
        args: the arguments that are being used
        length: the length (in bits)

    AST objects have *hash identity*. This means that an AST that has the same hash as
    another AST will be the *same* object. For example, the following is true:

        a, b = two different ASTs
        c = b + a
        d = b + a
        assert c is d

    This is done to better support serialization and better manage memory.
    '''

    __slots__ = [ 'op', 'args', 'variables', 'symbolic', '_hash', '_simplified',
                  '_cache_key', '_errored', '_eager_backends', 'length', '_excavated', '_burrowed', '_uninitialized',
                  '_uc_alloc_depth']
    _hash_cache = weakref.WeakValueDictionary()

    FULL_SIMPLIFY=1
    LITE_SIMPLIFY=2
    UNSIMPLIFIED=0

    def __new__(cls, op, args, **kwargs):
        '''
        This is called when you create a new Base object, whether directly or through an operation.
        It finalizes the arguments (see the _finalize function, above) and then computes
        a hash. If an AST of this hash already exists, it returns that AST. Otherwise,
        it creates, initializes, and returns the AST.

        @param op: the AST operation ('__add__', 'Or', etc)
        @param args: the arguments to the AST operation (i.e., the objects to add)
        @param variables: the symbolic variables present in the AST (default: empty set)
        @param symbolic: a flag saying whether or not the AST is symbolic (default: False)
        @param length: an integer specifying the length of this AST (default: None)
        @param collapsible: a flag of whether or not Claripy can feel free to collapse this AST.
                            This is mostly used to keep Claripy from collapsing Reverse operations,
                            so that they can be undone with another Reverse.
        @param simplified: a measure of how simplified this AST is. 0 means unsimplified, 1 means
                           fast-simplified (basically, just undoing the Reverse op), and 2 means
                           simplified through z3.
        @param errored: a set of backends that are known to be unable to handle this AST.
        @param eager_backends: a list of backends with which to attempt eager evaluation
        '''

        #if any(isinstance(a, BackendObject) for a in args):
        #   raise Exception('asdf')

        # fix up args and kwargs
        a_args = tuple((a.to_claripy() if isinstance(a, BackendObject) else a) for a in args)
        if 'symbolic' not in kwargs:
            kwargs['symbolic'] = any(a.symbolic for a in a_args if isinstance(a, Base))
        if 'variables' not in kwargs:
            kwargs['variables'] = frozenset.union(
                frozenset(), *(a.variables for a in a_args if isinstance(a, Base))
            )
        elif type(kwargs['variables']) is not frozenset: #pylint:disable=unidiomatic-typecheck
            kwargs['variables'] = frozenset(kwargs['variables'])
        if 'errored' not in kwargs:
            kwargs['errored'] = set.union(set(), *(a._errored for a in a_args if isinstance(a, Base)))

        if 'add_variables' in kwargs:
            kwargs['variables'] = kwargs['variables'] | kwargs['add_variables']

        eager_backends = list(backends._eager_backends) if 'eager_backends' not in kwargs else kwargs['eager_backends']

        if not kwargs['symbolic'] and eager_backends is not None and op not in operations.leaf_operations:
            for eb in eager_backends:
                try:
                    return eb._abstract(eb.call(op, args))
                except BackendError:
                    eager_backends.remove(eb)

        # if we can't be eager anymore, null out the eagerness
        kwargs['eager_backends'] = None
        h = Base._calc_hash(op, a_args, kwargs)

        # whether this guy is initialized or not
        if 'uninitialized' not in kwargs:
            kwargs['uninitialized'] = None

        if 'uc_alloc_depth' not in kwargs:
            kwargs['uc_alloc_depth'] = None

        self = cls._hash_cache.get(h, None)
        if self is None:
            self = super(Base, cls).__new__(cls, op, a_args, **kwargs)
            self.__a_init__(op, a_args, **kwargs)
            self._hash = h
            cls._hash_cache[h] = self
        # else:
        #    if self.args != f_args or self.op != f_op or self.variables != f_kwargs['variables']:
        #        raise Exception("CRAP -- hash collision")

        return self

    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def _calc_hash(op, args, k):
        '''
        Calculates the hash of an AST, given the operation, args, and kwargs.

        @param op: the operation
        @param args: the arguments to the operation
        @param kwargs: a dict including the 'symbolic', 'variables', and 'length' items

        @returns a hash
        '''
        to_hash = (op, tuple(str(a) if type(a) in (int, long) else hash(a) for a in args), k['symbolic'], hash(k['variables']), str(k.get('length', None)))
        # Why do we use md5 when it's broken? Because speed is more important
        # than cryptographic integrity here. Then again, look at all those
        # allocations we're doing here... fast python is painful.
        hd = hashlib.md5(pickle.dumps(to_hash, -1)).digest()
        return md5_unpacker.unpack(hd)[0] # 64 bits

    def _get_hashables(self):
        return self.op, tuple(str(a) if type(a) in (int, long) else hash(a) for a in self.args), self.symbolic, hash(self.variables), str(self.length)

    #pylint:disable=attribute-defined-outside-init
    def __a_init__(self, op, args, variables=None, symbolic=None, length=None, collapsible=None, simplified=0, errored=None, eager_backends=None, add_variables=None, uninitialized=None, uc_alloc_depth=None): #pylint:disable=unused-argument
        '''
        Initializes an AST. Takes the same arguments as Base.__new__()
        '''
        self.op = op
        self.args = args
        self.length = length
        self.variables = frozenset(variables)
        self.symbolic = symbolic
        self._eager_backends = eager_backends

        self._errored = errored if errored is not None else set()

        self._simplified = simplified
        self._cache_key = ASTCacheKey(self)
        self._excavated = None
        self._burrowed = None

        self._uninitialized = uninitialized
        self._uc_alloc_depth = uc_alloc_depth

        if len(args) == 0:
            raise ClaripyOperationError("AST with no arguments!")

        #if self.op != 'I':
        #    for a in args:
        #        if not isinstance(a, Base) and type(a) not in (int, long, bool, str, unicode):
        #            import ipdb; ipdb.set_trace()
        #            l.warning(ClaripyOperationError("Un-wrapped native object of type %s!" % type(a)))
    #pylint:enable=attribute-defined-outside-init

    def make_uuid(self, uuid=None):
        '''
        This overrides the default ANA uuid with the hash of the AST. UUID is slow,
        and we'll soon replace it from ANA itself, and this will go away.

        @returns a string representation of the AST hash.
        '''
        u = getattr(self, '_ana_uuid', None)
        if u is None:
            u = str(self._hash) if uuid is None else uuid
            ana.get_dl().uuid_cache[u] = self
            setattr(self, '_ana_uuid', u)
        return u

    @property
    def uuid(self):
        return self.ana_uuid

    def __hash__(self):
        return self._hash

    @property
    def cache_key(self):
        return self._cache_key

    #
    # Serialization support
    #

    def _ana_getstate(self):
        '''
        Support for ANA serialization.
        '''
        return self.op, self.args, self.length, self.variables, self.symbolic, self._hash
    def _ana_setstate(self, state):
        '''
        Support for ANA deserialization.
        '''
        op, args, length, variables, symbolic, h = state
        Base.__a_init__(self, op, args, length=length, variables=variables, symbolic=symbolic)
        self._hash = h
        Base._hash_cache[h] = self

    #
    # Collapsing and simplification
    #

    #def _models_for(self, backend):
    #    for a in self.args:
    #        backend.convert_expr(a)
    #        else:
    #            yield backend.convert(a)

    def make_like(self, *args, **kwargs):
        return type(self)(*args, **kwargs)

    #
    # Viewing and debugging
    #

    def dbg_repr(self, prefix=None):
        try:
            if prefix is not None:
                new_prefix = prefix + "    "
                s = prefix + "<%s %s (\n" % (type(self).__name__, self.op)
                for a in self.args:
                    s += "%s,\n" % (a.dbg_repr(prefix=new_prefix) if hasattr(a, 'dbg_repr') else (new_prefix + repr(a)))
                s = s[:-2] + '\n'
                s += prefix + ")>"

                return s
            else:
                return "<%s %s (%s)>" % (type(self).__name__, self.op, ', '.join(a.dbg_repr() if hasattr(a, 'dbg_repr') else repr(a) for a in self.args))
        except RuntimeError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyRecursionError, ("Recursion limit reached during display. I sorry.", e_type, value), traceback

    def _type_name(self):
        return self.__class__.__name__

    def __repr__(self, inner=False, explicit_length=False):
        if WORKER:
            return '<AST something>'

        try:
            if self.op in operations.reversed_ops:
                op = operations.reversed_ops[self.op]
                args = self.args[::-1]
            else:
                op = self.op
                args = self.args

            if op == 'BVS' and inner:
                value = args[0]
            elif op == 'BoolV':
                value = str(args[0])
            elif op == 'BVV':
                if self.args[0] is None:
                    value = '!'
                elif self.args[1] < 10:
                    value = format(self.args[0], '')
                else:
                    value = format(self.args[0], '#x')
                value += ('#' + str(self.length)) if explicit_length else ''
            elif op == 'If':
                value = 'if {} then {} else {}'.format(_inner_repr(args[0]),
                                                       _inner_repr(args[1]),
                                                       _inner_repr(args[2]))
                if inner:
                    value = '({})'.format(value)
            elif op == 'Not':
                value = '!{}'.format(_inner_repr(args[0]))
            elif op == 'Extract':
                value = '{}[{}:{}]'.format(_inner_repr(args[2]), args[0], args[1])
            elif op == 'ZeroExt':
                value = '0#{} .. {}'.format(args[0], _inner_repr(args[1]))
                if inner:
                    value = '({})'.format(value)
            elif op == 'Concat':
                value = ' .. '.join(_inner_repr(a, explicit_length=True) for a in self.args)
            elif len(args) == 2 and op in operations.infix:
                value = '{} {} {}'.format(_inner_repr(args[0]),
                                          operations.infix[op],
                                          _inner_repr(args[1]))
                if inner:
                    value = '({})'.format(value)
            else:
                value = "{}({})".format(op,
                                        ', '.join(_inner_repr(a) for a in args))

            if not inner:
                value = '<{} {}>'.format(self._type_name(), value)

            return value
        except RuntimeError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyRecursionError, ("Recursion limit reached during display. I sorry.", e_type, value), traceback

    @property
    def depth(self):
        '''
        The depth of this AST. For example, an AST representing (a+(b+c)) would have
        a depth of 2.
        '''
        return self._depth()

    def _depth(self, memoized=None):
        """
        :param memoized: dict of ast hashes to depths we've seen before
        :return: the depth of the AST. For example, an AST representing (a+(b+c)) would have
        a depth of 2.
        """
        if memoized is None:
            memoized = dict()

        ast_args = [ a for a in self.args if isinstance(a, Base) ]
        max_depth = 0
        for a in ast_args:
            if a not in memoized:
                memoized[a] = a._depth(memoized)
            max_depth = max(memoized[a], max_depth)

        return 1 + max_depth

    @property
    def recursive_children_asts(self):
        for a in self.args:
            if isinstance(a, Base):
                l.debug("Yielding AST %s with hash %s with %d children", a, hash(a), len(a.args))
                yield a
                for b in a.recursive_children_asts:
                    yield b

    @property
    def recursive_leaf_asts(self):
        for a in self.args:
            if isinstance(a, Base):
                if a.op in ('BVS', 'BVV', 'I'):
                    yield a
                else:
                    for b in a.recursive_leaf_asts:
                        yield b

    def dbg_is_looped(self, seen=None, checked=None):
        seen = set() if seen is None else seen
        checked = set() if checked is None else checked

        l.debug("Checking AST with hash %s for looping", hash(self))
        if hash(self) in seen:
            return self
        elif hash(self) in checked:
            return False
        else:
            seen.add(hash(self))

            for a in self.args:
                if not isinstance(a, Base):
                    continue

                r = a.dbg_is_looped(seen=set(seen), checked=checked)
                if r is not False:
                    return r

            checked.add(hash(self))
            return False

    #
    # Various AST modifications (replacements)
    #

    def _replace(self, replacements, variable_set=None):
        """
        A helper for replace().
        :param variable_set: for optimization, ast's without these variables are not checked for replacing
        :param replacements: dictionary of hashes to their replacements
        """
        if variable_set is None:
            variable_set = {}

        hash_key = self.cache_key

        if hash_key in replacements:
            r = replacements[hash_key]
        elif not self.variables.issuperset(variable_set):
            r = self
        else:
            new_args = [ ]
            replaced = False

            for a in self.args:
                if isinstance(a, Base):
                    new_a = a._replace(replacements=replacements, variable_set=variable_set)
                    replaced |= hash(new_a) != hash(a)
                else:
                    new_a = a

                new_args.append(new_a)

            if replaced:
                r = self.make_like(self.op, tuple(new_args))
            else:
                r = self

        replacements[hash_key] = r
        return r

    def swap_args(self, new_args, new_length=None):
        '''
        This returns the same AST, with the arguments swapped out for new_args.
        '''

        if len(self.args) == len(new_args) and all(a is b for a,b in zip(self.args, new_args)):
            return self

        #symbolic = any(a.symbolic for a in new_args if isinstance(a, Base))
        #variables = frozenset.union(frozenset(), *(a.variables for a in new_args if isinstance(a, Base)))
        length = self.length if new_length is None else new_length
        a = self.__class__(self.op, new_args, length=length)
        #if a.op != self.op or a.symbolic != self.symbolic or a.variables != self.variables:
        #   raise ClaripyOperationError("major bug in swap_args()")
        return a

    #
    # Other helper functions
    #

    def split(self, split_on):
        '''
        Splits the AST if its operation is split_on (i.e., return all the arguments).
        Otherwise, return a list with just the AST.
        '''
        if self.op in split_on: return list(self.args)
        else: return [ self ]

    # we don't support iterating over Base objects
    def __iter__(self):
        '''
        This prevents people from iterating over ASTs.
        '''
        raise ClaripyOperationError("Please don't iterate over, or split, AST nodes!")

    def __nonzero__(self):
        '''
        This prevents people from accidentally using an AST as a condition. For
        example, the following was previously common:

            a,b = two ASTs
            if a == b:
                do something

        The problem is that `a == b` would return an AST, because an AST can be symbolic
        and there could be no way to actually know the value of that without a
        constraint solve. This caused tons of issues.
        '''
        raise ClaripyOperationError('testing Expressions for truthiness does not do what you want, as these expressions can be symbolic')

    def structurally_match(self, o):
        """
        Structurally compares two A objects, and check if their corresponding leaves are definitely the same A object
        (name-wise or hash-identity wise).

        :param o: the other claripy A object
        :return: True/False
        """

        # TODO: Convert a and b into canonical forms

        if self.op != o.op:
            return False

        if len(self.args) != len(o.args):
            return False

        for arg_a, arg_b in zip(self.args, o.args):
            if not isinstance(arg_a, Base):
                if type(arg_a) != type(arg_b):
                    return False
                # They are not ASTs
                if arg_a != arg_b:
                    return False
                else:
                    continue

            if arg_a.op in ('I', 'BVS', 'FP'):
                # This is a leaf node in AST tree
                if arg_a is not arg_b:
                    return False

            else:
                if not arg_a.structurally_match(arg_b):
                    return False

        return True

    def replace(self, old, new):
        '''
        Returns an AST with all instances of the AST 'old' replaced with AST 'new'
        '''
        if not isinstance(old, Base) or not isinstance(new, Base):
            raise ClaripyOperationError('replacements must be AST nodes')
        if type(old) is not type(new):
            raise ClaripyOperationError('cannot replace type %s ast with type %s ast' % (type(old), type(new)))
        old._check_replaceability(new)
        replacements = {old.cache_key: new}
        return self._replace(replacements, old.variables)

    def replace_dict(self, replacements):
        """
        :param replacements: a dictionary of asts to replace and their replacements
        :return: an AST with all instances of ast's in replacements
        """
        for old, new in replacements.items():
            old = old.ast
            if not isinstance(old, Base) or not isinstance(new, Base):
                raise ClaripyOperationError('replacements must be AST nodes')
            if type(old) is not type(new):
                raise ClaripyOperationError('cannot replace type %s ast with type %s ast' % (type(old), type(new)))
            old._check_replaceability(new)

        return self._replace(replacements)

    def _check_replaceability(self, new):
        pass

    def _identify_vars(self, all_vars, counter):
        if self.op == 'BVS':
            if self.args not in all_vars:
                all_vars[self.args] = BV('var_' + str(next(counter)),
                                         self.args[1],
                                         explicit_name=True)
        else:
            for arg in self.args:
                if isinstance(arg, Base):
                    arg._identify_vars(all_vars, counter)

    def canonicalized(self, existing_vars=None, counter=None):
        all_vars = {} if existing_vars is None else existing_vars
        counter = itertools.count() if counter is None else counter
        self._identify_vars(all_vars, counter)

        expr = self
        for old_var, new_var in all_vars.items():
            expr = expr.replace(BV(*old_var, explicit_name=True), new_var)

        return all_vars, expr

    #
    # This code handles burrowing ITEs deeper into the ast and excavating
    # them to shallower levels.
    #

    def _burrow_ite(self):
        if self.op != 'If':
            #print "i'm not an if"
            return self.swap_args([ (a.ite_burrowed if isinstance(a, Base) else a) for a in self.args ])

        if not all(isinstance(a, Base) for a in self.args):
            #print "not all my args are bases"
            return self

        old_true = self.args[1]
        old_false = self.args[2]

        if old_true.op != old_false.op or len(old_true.args) != len(old_false.args):
            return self

        if old_true.op == 'If':
            # let's no go into this right now
            return self

        if any(a.op in {'BVS', 'BVV', 'FPS', 'FPV', 'BoolS', 'BoolV'} for a in self.args):
            # burrowing through these is pretty funny
            return self

        matches = [ old_true.args[i] is old_false.args[i] for i in range(len(old_true.args)) ]
        if matches.count(True) != 1 or all(matches):
            # TODO: handle multiple differences for multi-arg ast nodes
            #print "wrong number of matches:",matches,old_true,old_false
            return self

        different_idx = matches.index(False)
        inner_if = If(self.args[0], old_true.args[different_idx], old_false.args[different_idx])
        new_args = list(old_true.args)
        new_args[different_idx] = inner_if.ite_burrowed
        #print "replaced the",different_idx,"arg:",new_args
        return old_true.__class__(old_true.op, new_args, length=self.length)

    def _excavate_ite(self):
        if self.op in { 'BVS', 'I', 'BVV' }:
            return self

        excavated_args = [ (a.ite_excavated if isinstance(a, Base) else a) for a in self.args ]
        ite_args = [ isinstance(a, Base) and a.op == 'If' for a in excavated_args ]

        if self.op == 'If':
            # if we are an If, call the If handler so that we can take advantage of its simplifiers
            return If(*excavated_args)
        elif ite_args.count(True) == 0:
            # if there are no ifs that came to the surface, there's nothing more to do
            return self.swap_args(excavated_args)
        else:
            # this gets called when we're *not* in an If, but there are Ifs in the args.
            # it pulls those Ifs out to the surface.
            cond = excavated_args[ite_args.index(True)].args[0]
            new_true_args = [ ]
            new_false_args = [ ]

            for a in excavated_args:
                #print "OC", cond.dbg_repr()
                #print "NC", Not(cond).dbg_repr()

                if not isinstance(a, Base) or a.op != 'If':
                    new_true_args.append(a)
                    new_false_args.append(a)
                elif a.args[0] is cond:
                    #print "AC", a.args[0].dbg_repr()
                    new_true_args.append(a.args[1])
                    new_false_args.append(a.args[2])
                elif a.args[0] is Not(cond):
                    #print "AN", a.args[0].dbg_repr()
                    new_true_args.append(a.args[2])
                    new_false_args.append(a.args[1])
                else:
                    #print "AB", a.args[0].dbg_repr()
                    # weird conditions -- giving up!
                    return self.swap_args(excavated_args)

            return If(cond, self.swap_args(new_true_args), self.swap_args(new_false_args))

    @property
    def ite_burrowed(self):
        '''
        Returns an equivalent AST that "burrows" the ITE expressions
        as deep as possible into the ast, for simpler printing.
        '''
        if self._burrowed is None:
            self._burrowed = self._burrow_ite() #pylint:disable=attribute-defined-outside-init
            self._burrowed._burrowed = self._burrowed
        return self._burrowed

    @property
    def ite_excavated(self):
        '''
        Returns an equivalent AST that "excavates" the ITE expressions
        out as far as possible toward the root of the AST, for processing
        in static analyses.
        '''
        if self._excavated is None:
            self._excavated = self._excavate_ite() #pylint:disable=attribute-defined-outside-init

            # we set the flag for the children so that we avoid re-excavating during
            # VSA backend evaluation (since the backend evaluation recursively works on
            # the excavated ASTs)
            self._excavated._excavated = self._excavated
        return self._excavated

    #
    # these are convenience operations
    #

    def _first_backend(self, what):
        for b in backends._all_backends:
            if b in self._errored:
                continue

            try: return getattr(b, what)(self)
            except BackendError: pass

    @property
    def singlevalued(self):
        return self._first_backend('singlevalued')

    @property
    def multivalued(self):
        return self._first_backend('multivalued')

    @property
    def cardinality(self):
        return self._first_backend('cardinality')

    @property
    def concrete(self):
        return backends.concrete.handles(self)

    @property
    def uninitialized(self):
        """
        Whether this AST comes from an uninitialized dereference or not. It's only used in under-constrained symbolic execution
        mode.

        :return: True/False/None (unspecified)
        """

        #TODO: It should definitely be moved to the proposed Annotation backend.

        return self._uninitialized

    @property
    def uc_alloc_depth(self):
        """
        The depth of allocation by lazy-initialization. It's only used in under-constrained symbolic execution mode.

        :return: An integer indicating the allocation depth, or None if it's not from lazy-initialization.
        """
        # TODO: It should definitely be moved to the proposed Annotation backend.

        return self._uc_alloc_depth

    #
    # Backwards compatibility crap
    #

    @property
    def model(self):
        l.critical("DEPRECATION WARNING: do not use AST.model. It is deprecated, no longer does what is expected, and will soon be removed. If you *need* to access the model use AST._model_X where X is the backend that you are interested in.")
        print "DEPRECATION WARNING: do not use AST.model. It is deprecated, no longer does what is expected, and will soon be removed. If you *need* to access the model use AST._model_X where X is the backend that you are interested in."
        return self._model_concrete if self._model_concrete is not self else \
               self._model_vsa if self._model_vsa is not self else \
               self._model_z3 if self._model_z3 is not self else \
               self

    def __getattr__(self, a):
        if not a.startswith('_model_'):
            raise AttributeError(a)

        model_name = a[7:]
        if not hasattr(backends, model_name):
            raise AttributeError(a)

        try:
            return getattr(backends, model_name).convert(self)
        except BackendError:
            return self

def simplify(e):
    if isinstance(e, Base) and e.op == 'I':
        return e

    s = e._first_backend('simplify')
    if s is None:
        l.debug("Unable to simplify expression")
        return e
    else:
        # Copy some parameters (that should really go to the Annotation backend)
        s._uninitialized = e.uninitialized
        s._uc_alloc_depth = e._uc_alloc_depth

        s._simplified = Base.FULL_SIMPLIFY

        return s

from ..errors import BackendError, ClaripyOperationError, ClaripyRecursionError
from .. import operations
from ..backend_object import BackendObject
from ..backend_manager import backends
from ..ast.bool import If, Not
from ..ast.bv import BV



import cPickle as pickle
import functools
import itertools
import logging
import time
l = logging.getLogger('claripy.backends.remotetasks')

from . import celeryconfig

import celery
from celery.contrib import rdb
import pymongo
import bson

app = celery.Celery('tasks', broker='amqp://guest@localhost//', backend='mongodb://localhost/')
app.config_from_object(celeryconfig)
z3 = None
mongo = None

@celery.signals.celeryd_after_setup.connect
def init_z3_and_conns(sender, instance, **kwargs):
    global z3, mongo
    z3 = claripy.backends.BackendZ3()
    z3.set_claripy_object(claripy.Claripies['SerialZ3'])
    mongo = pymongo.MongoClient()['lemma_cache']

def flip_dict(d):
    return {val: key for (key, val) in d.items()}

def canonicalize_all(exprs):
    known_vars = {}
    counter = itertools.count()
    return known_vars, [expr.canonicalized(existing_vars=known_vars, counter=counter)[1] for expr in exprs]

def split_and_canonicalize(constraints):
    independent = Solver.independent_constraints(constraints)
    # TODO: deal with different *orders* of constraints
    parts = {}
    for _, cset in independent:
        mapping, all_canonicalized = canonicalize_all(cset)
        # note: the `sorted` does not fix the above TODO because it
        # does not account for different orderings in variable
        # renamings
        parts[tuple(sorted(str(e._hash) for e in all_canonicalized))] = (mapping, all_canonicalized)
    return parts

@app.task
def check(constraints):
    return results(constraints).sat
    # global z3

    # parts = split_and_canonicalize(constraints)
    # all_uuids = sorted(uu for uuids in parts for uu in uuids)

    # if mongo['results'].find({'uuids': {'$in': all_uuids}, 'sat': False}).limit(1).count(with_limit_and_skip=True) > 0:
    #     # this means a constraint set consisting of a subset of the
    #     # uuids is unsat, meaning all the uuids together must also be
    #     # unsat
    #     return False

    # for doc in mongo['results'].find({'$or': [{'uuids': uu} for uu in parts]}):
    #     l.info("lemma cache find, result for %r is %s", doc['uuids'], doc['sat'])
    #     if doc['sat']:
    #         # this part is sat, don't need to check it again
    #         del parts[tuple(doc['uuids'])]
    #     else:
    #         # this part is unsat, so as a whole, constraints must be unsat
    #         return False

    # s = z3.solver()

    # for uuids, (_, exprs) in parts.items():
    #     sat = z3.check(s, extra_constraints=exprs)
    #     mongo['results'].insert_one({'uuids': uuids, 'sat': sat, 'model': None})
    #     if not sat:
    #         return False

    # return True

@app.task
def eval(constraints, expr, n):
    global z3
    s = z3.solver()
    z3.add(s, constraints)
    return z3.eval(s, expr, n)

@app.task
def batch_eval(constraints, exprs, n):
    global z3
    s = z3.solver()
    z3.add(s, constraints)
    return z3.batch_eval(s, exprs, n)

def rename_model(mapping, model, transform=lambda x: x):
    print model
    return {
        # if the name isn't in the model, then it's a "don't care",
        # i.e., it can hold any value
        orig_var_name: transform(model[generic_var.args[0]]) if generic_var.args[0] in model else generic_var._claripy.BVV(0, sz)
        for ((orig_var_name, sz), generic_var)
        in mapping.items()
    }

def report_sat(f):
    @functools.wraps(f)
    def wrapped(constraints):
        result = f(constraints)
        l.warning("result.sat = %s", result.sat)
        return result
    return wrapped

def mongo_superset(s):
    # this is probably highly inefficient...
    return {'$not': {'$elemMatch': {'$nin': s}}}

@app.task
@report_sat
def results(constraints):
    global z3

    with open('./foo.p', 'wb') as f:
        pickle.dump(constraints, f)

    if len(constraints) == 0:
        return Result(True, {})

    before = time.time()

    model = {}
    parts = split_and_canonicalize(constraints)
    all_uuids = sorted(uu for uuids in parts for uu in uuids)

    print "before mongo stuff took %f" % (time.time() - before)

    if mongo['results'].find({'uuids': mongo_superset(all_uuids), 'sat': False}).limit(1).count(with_limit_and_skip=True) > 0:
        # this means a constraint set consisting of a subset of the
        # uuids is unsat, meaning all the uuids together must also be
        # unsat
        print "found unsat $in"
        return Result(False, {})

    print "whole pre-solve took %f" % (time.time() - before)

    # for doc in mongo['results'].find({'$or': [{'uuids': uu} for uu in parts]}):
    #     l.info("lemma cache find for %r, sat is %s and model is %s", doc['uuids'], doc['sat'], doc['model'])
    #     # if doc['sat']:
    #     #     if doc['model'] is not None:
    #     #         mapping, _ = parts[tuple(doc['uuids'])]
    #     #         model.update(rename_model(mapping, doc['model'], transform=pickle.loads))
    #     #         del parts[tuple(doc['uuids'])]
    #     if not doc['sat']:
    #         return Result(False, {})

    s = z3.solver()

    for uuids, (mapping, exprs) in parts.items():
        if all(e.is_true() for e in exprs):
            continue

        result = z3.results(s, extra_constraints=exprs, generic_model=True)
        doc = {
            'uuids': uuids,
            'sat': result.sat,
            'model': None,
        }
        if result.sat:
            doc['model'] = {name: bson.binary.Binary(pickle.dumps(val, protocol=pickle.HIGHEST_PROTOCOL))
                            for (name, val) in result.model.items()}
            try:
                model.update(rename_model(mapping, result.model))
            except KeyError:
                rdb.set_trace()

        mongo['results'].replace_one({'uuids': uuids}, doc, upsert=True)

        if not result.sat:
            return Result(False, {})

    return Result(True, model)

@app.task
def min(constraints, expr):
    global z3
    s = z3.solver()
    z3.add(s, constraints)
    return z3.min(s, expr)

@app.task
def max(constraints, expr):
    global z3
    s = z3.solver()
    z3.add(s, constraints)
    return z3.max(s, expr)


import claripy
from ..solvers import Solver
from ..result import Result
from ..ast import base



import functools
import itertools
from ..backend_object import BackendObject

def normalize_types(f):
    @functools.wraps(f)
    def normalizer(self, region, o):
        '''
        Convert any object to an object that we can process.
        '''
        if isinstance(o, Base):
            raise ClaripyValueError("BoolResult can't handle AST objects directly")

        if not isinstance(o, StridedInterval):
            raise ClaripyVSAOperationError('Unsupported operand type %s' % type(o))

        return f(self, region, o)

    return normalizer

def normalize_types_one_arg(f):
    @functools.wraps(f)
    def normalizer(self, o):
        '''
        Convert any object to an object that we can process.
        '''
        if isinstance(o, Base):
            raise ClaripyValueError("BoolResult can't handle AST objects directly")

        return f(self, o)

    return normalizer

vs_id_ctr = itertools.count()

class ValueSet(BackendObject):
    def __init__(self, name=None, region=None, bits=None, val=None):
        self._name = name
        if self._name is None:
            self._name = 'VS_%d' % vs_id_ctr.next()

        self._regions = {}

        self._reversed = False
        if bits is None:
            if val is None:
                raise ClaripyVSAError("Either 'size' or 'val' must be specified.")
            else:
                self._bits = val.bits()

        else:
            self._bits = bits

        if region is not None and val is not None:
            if isinstance(val, StridedInterval):
                self.set_si(region, val)

            elif type(val) in (int, long):
                self.set_si(region, StridedInterval(bits=bits, stride=0, lower_bound=val, upper_bound=val))

            else:
                raise ClaripyVSAError("Unsuported type '%s' for argument 'val'" % type(val))

        else:
            if region is not None or val is not None:
                raise ClaripyVSAError("You must specify 'region' and 'val' at the same time.")

    @property
    def name(self):
        return self._name

    @property
    def regions(self):
        return self._regions

    @property
    def reversed(self):
        return self._reversed

    @property
    def unique(self):
        return len(self.regions) == 1 and self.regions.values()[0].unique

    @property
    def bits(self):
        return self.size()

    @property
    def cardinality(self):
        card = 0
        for si in self._regions.itervalues():
            card += si.cardinality

        return card

    @normalize_types
    def set_si(self, region, si):
        if not isinstance(si, StridedInterval):
            raise ClaripyVSAOperationError('Unsupported type %s for si' % type(si))

        self._regions[region] = si

    def get_si(self, region):
        if region in self._regions:
            return self._regions[region]
        # TODO: Should we return a None, or an empty SI instead?
        return None

    def items(self):
        return self._regions.items()

    def size(self):
        return len(self)

    @normalize_types
    def merge_si(self, region, si):
        if region not in self._regions:
            self.set_si(region, si)
        else:
            self._regions[region] = self._regions[region].union(si)

    @normalize_types
    def remove_si(self, region, si):
        raise NotImplementedError()

    def __repr__(self):
        s = ""
        for region, si in self._regions.items():
            s = "%s: %s" % (region, si)
        return "(" + s + ")"

    def __len__(self):
        return self._bits

    @normalize_types_one_arg
    def __add__(self, other):
        if type(other) is ValueSet:
            # Normally, addition between two addresses doesn't make any sense.
            # So we only handle those corner cases

            raise NotImplementedError()
        else:
            new_vs = ValueSet(bits=self.bits)
            for region, si in self._regions.items():
                new_vs._regions[region] = si + other

            return new_vs

    @normalize_types_one_arg
    def __radd__(self, other):
        return self.__add__(other)

    @normalize_types_one_arg
    def __sub__(self, other):
        """
        Binary operation: subtraction

        :param other: The other operand
        :return: A StridedInterval.
        """

        deltas = [ ]

        # TODO: Handle more cases

        if type(other) is ValueSet:
            # A subtraction between two ValueSets produces a StridedInterval

            if self.regions.keys() == other.regions.keys():
                for region, si in self._regions.iteritems():
                    deltas.append(si - other._regions[region])

            else:
                __import__('ipdb').set_trace()
                raise NotImplementedError()

            delta = StridedInterval.empty(self.bits)
            for d in deltas:
                delta = delta.union(d)

            return delta

        else:
            # A subtraction between a ValueSet and an StridedInterval produces another ValueSet

            new_vs = self.copy()
            for region, si in new_vs._regions.items():
                new_vs._regions[region] = si - other

            return new_vs

    @normalize_types_one_arg
    def __and__(self, other):
        if type(other) is ValueSet:
            if self.identical(other):
                return self.copy()
            raise NotImplementedError("__and__(vs1, vs2)?")

        if BoolResult.is_true(other == 0):
            # Corner case: a & 0 = 0
            return StridedInterval(bits=self.bits, stride=0, lower_bound=0, upper_bound=0)

        new_vs = ValueSet(bits=self.bits)
        if BoolResult.is_true(other < 0x100):
            # Special case - sometimes (addr & mask) is used for testing whether the address is aligned or not
            # We return an SI instead
            ret = None

            for region, si in self._regions.items():
                r = si.__and__(other)

                ret = r if ret is None else ret.union(r)

            return ret

        else:
            for region, si in self._regions.items():
                r = si.__and__(other)

                new_vs._regions[region] = r

            return new_vs

    def __eq__(self, other):
        if isinstance(other, ValueSet):
            same = False
            different = False
            for region, si in other.regions.items():
                if region in self.regions:
                    comp_ret = self.regions[region] == si
                    if BoolResult.has_true(comp_ret):
                        same = True
                    if BoolResult.has_false(comp_ret):
                        different = True
                else:
                    different = True

            if same and not different:
                return TrueResult()
            if same and different:
                return MaybeResult()
            return FalseResult()
        elif isinstance(other, StridedInterval):
            if 'global' in self.regions:
                return self.regions['global'] == other
            else:
                return FalseResult()
        else:
            return FalseResult()

    def __ne__(self, other):
        return ~ (self == other)

    def eval(self, n):
        results = []

        for _, si in self._regions.items():
            if len(results) < n:
                results.extend(si.eval(n))

        return results

    def copy(self):
        vs = ValueSet(bits=self.bits)
        vs._regions = self._regions.copy()
        vs._reversed = self._reversed

        return vs

    def reverse(self):
        # TODO: obviously valueset.reverse is not properly implemented. I'm disabling the old annoying output line for
        # TODO: now. I will implement the proper reversing support soon.
        vs = self.copy()
        vs._reversed = not vs._reversed

        return vs

    @property
    def is_empty(self):
        return len(self._regions) == 0

    def extract(self, high_bit, low_bit):
        """
        Operation extract

        - A cheap hack is implemented: a copy of self is returned if (high_bit - low_bit + 1 == self.bits), which is a
            ValueSet instance. Otherwise a StridedInterval is returned.

        :param high_bit:
        :param low_bit:
        :return: A ValueSet or a StridedInterval
        """

        if high_bit - low_bit + 1 == self.bits:
            return self.copy()

        if ('global' in self._regions and len(self._regions.keys()) > 1) or \
            len(self._regions.keys()) > 0:
            si_ret = StridedInterval.top(high_bit - low_bit + 1)

        else:
            if 'global' in self._regions:
                si = self._regions['global']
                si_ret = si.extract(high_bit, low_bit)

            else:
                si_ret = StridedInterval.empty(high_bit - low_bit + 1)

        return si_ret

    def concat(self, b):
        new_vs = ValueSet(bits=self.bits + b.bits)
        # TODO: This logic is obviously flawed. Correct it later :-(
        if isinstance(b, StridedInterval):
            for region, si in self._regions.items():
                new_vs.set_si(region, si.concat(b))

        elif isinstance(b, ValueSet):
            for region, si in self._regions.items():
                new_vs.set_si(region, si.concat(b.get_si(region)))

        else:
            raise ClaripyVSAOperationError('ValueSet.concat() got an unsupported operand %s (type %s)' % (b, type(b)))

        return new_vs

    @normalize_types_one_arg
    def union(self, b):
        merged_vs = self.copy()
        if type(b) is ValueSet:
            for region, si in b.regions.items():
                if region not in merged_vs._regions:
                    merged_vs._regions[region] = si
                else:
                    merged_vs._regions[region] = merged_vs._regions[region].union(si)
        else:
            for region, si in self._regions.items():
                merged_vs._regions[region] = merged_vs._regions[region].union(b)

        return merged_vs

    @normalize_types_one_arg
    def widen(self, b):
        merged_vs = self.copy()
        if isinstance(b, ValueSet):
            for region, si in b.regions.items():
                if region not in merged_vs.regions:
                    merged_vs.regions[region] = si
                else:
                    merged_vs.regions[region] = merged_vs.regions[region].widen(si)

        else:
            for region, si in self._regions.iteritems():
                merged_vs._regions[region] = merged_vs._regions[region].widen(b)

        return merged_vs

    def identical(self, o):
        """
        Used to make exact comparisons between two ValueSets.

        :param o: The other ValueSet to compare with
        :return: True if they are exactly same, False otherwise
        """
        if self._reversed != o._reversed:
            return False

        for region, si in self.regions.items():
            if region in o.regions:
                o_si = o.regions[region]
                if not si.identical(o_si):
                    return False
            else:
                return False

        return True

    def __hash__(self):
        return hash(tuple([ (r, hash(si)) for r, si in self._regions.iteritems() ]))

from ..ast.base import Base
from .strided_interval import StridedInterval
from .bool_result import BoolResult, TrueResult, FalseResult, MaybeResult
from .errors import ClaripyVSAOperationError, ClaripyVSAError
from ..errors import ClaripyValueError



#!/usr/bin/env python

import logging
l = logging.getLogger("claripy.frontends.constrained_frontend")

from ..frontend import Frontend

class ConstrainedFrontend(Frontend): #pylint:disable=abstract-method
    def __init__(self, **kwargs):
        Frontend.__init__(self, **kwargs)
        self.constraints = [ ]
        self._constraint_hashes = set()
        self.variables = set()
        self._finalized = False

    #
    # Storable support
    #

    def _ana_getstate(self):
        if not self._simplified: self.simplify()
        self.finalize()
        return self.constraints, self.variables, Frontend._ana_getstate(self)

    def _ana_setstate(self, s):
        self.constraints, self.variables, base_state = s
        Frontend._ana_setstate(self, base_state)
        self._finalized = True

    #
    # Constraint management
    #

    def independent_constraints(self):
        return self._split_constraints(self.constraints)

    #
    # Light functionality
    #

    def _add_constraints(self, constraints, invalidate_cache=True):
        new_constraints = [ c for c in constraints if hash(c) not in self._constraint_hashes ]
        self.constraints += new_constraints
        for c in new_constraints:
            self.variables.update(c.variables)
            self._constraint_hashes.add(hash(c))
        return new_constraints


    def _simplify(self):
        if len(self.constraints) == 0:
            return

        self.constraints = [ simplify(And(*self.constraints)) ]

        # we only add to the constraint hashes because we want to
        # prevent previous (now simplified) constraints from
        # being re-added
        self._constraint_hashes.add(hash(self.constraints[0]))

        # generate UUIDs for every constraint
        for c in self.constraints:
            if isinstance(c, Base): c.make_uuid()

        self._simplified = True
        return self.constraints

    def _solve(self, extra_constraints=(), exact=None, cache=None):
        return SatResult(approximation=True)

    def _satisfiable(self, extra_constraints=(), exact=None, cache=None):
        return self.solve(extra_constraints=extra_constraints, exact=exact, cache=cache).sat

    #
    # Serialization and such.
    #

    def downsize(self):
        Frontend.downsize(self)

    #
    # Merging and splitting
    #

    def finalize(self):
        self._finalized = True

    def _blank_copy(self):
        return ConstrainedFrontend(cache=self._cache)

    def branch(self):
        s = Frontend.branch(self)
        s.constraints = list(self.constraints)
        s.variables = set(self.variables)
        s._constraint_hashes = set(self._constraint_hashes)
        self.finalize()
        s.finalize()
        return s

    def merge(self, others, merge_flag, merge_values):
        merged = self._blank_copy()
        merged._simplified = False
        options = [ ]

        for s, v in zip([self]+others, merge_values):
            options.append(And(*([ merge_flag == v ] + s.constraints)))
        merged.add([Or(*options)])

        return False, merged

    def combine(self, others):
        combined = self._blank_copy()
        combined._simplified = False

        combined.add(self.constraints) #pylint:disable=E1101
        for o in others:
            combined.add(o.constraints)
        return combined

    def split(self):
        results = [ ]
        l.debug("Splitting!")
        for variables,c_list in self.independent_constraints():
            l.debug("... got %d constraints with %d variables", len(c_list), len(variables))

            s = self._blank_copy()
            s._simplified = False
            s.add(c_list)
            results.append(s)
        return results

from ..result import SatResult
from ..ast.base import Base, simplify
from ..ast.bool import And, Or



class BackendObject(object):
    '''
    This is a base class for custom backend objects to implement.

    It lets Claripy know that how to deal with those objects, in case they're
    directly used in operations.

    Backend objects that *don't* derive from this class need to be wrapped in
    a type-I claripy.ast.Base.
    '''

    def to_claripy(self):
        '''
        Claripy calls this to retrieve something that it can directly reason
        about.
        '''

        return self




import weakref

class Replacer(object):
    def __init__(self, replacements=None, replacement_cache=None):
        self._replacements = { } if replacements is None else replacements
        self._replacement_cache = weakref.WeakKeyDictionary() if replacement_cache is None else replacement_cache

    def add_replacement(self, old, new, invalidate_cache=False):
        if invalidate_cache:
            self._replacements = dict(self._replacements)
            self._replacement_cache = weakref.WeakKeyDictionary(self._replacement_cache)

        self._replacements[old.cache_key] = new
        self._replacement_cache[old.cache_key] = new

    def replacement(self, old):
        if not isinstance(old, Base):
            return old

        if old.cache_key in self._replacement_cache:
            return self._replacement_cache[old.cache_key]
        else:
            new = old.replace_dict(self._replacement_cache)
            self._replacement_cache[old.cache_key] = new
            return new

    def branch(self):
        return Replacer(replacements=self._replacements, replacement_cache=self._replacement_cache)

    def __getstate__(self):
        return self._replacements

    def __setstate__(self, s):
        self._replacements = s
        self._replacement_cache = weakref.WeakKeyDictionary()

from .ast.base import Base



from ..ast.base import Base

class Bits(Base):
    __slots__ = ['length']

    def __init__(self, *args, **kwargs):
        length = kwargs.pop('length', None)
        if length is None:
            raise ClaripyOperationError("length of Bits must not be None")

        self.length = length

    def make_like(self, *args, **kwargs):
        kwargs['length'] = self.length
        return type(self)(*args, **kwargs)

    def size(self):
        return self.length

    def _type_name(self):
        return self.__class__.__name__ + str(self.length)

    def _check_replaceability(self, new):
        if self.size() != new.size():
            raise ClaripyOperationError('replacements must have matching sizes')

    __len__ = size


from ..errors import ClaripyOperationError
